# Copyright (c) 2016, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

import math
import cmath

import numpy as np
import quaternion
from quaternion.numba_wrapper import njit, jit, vectorize, int64, float64, xrange
from Wigner3j import clebsch_gordan


@vectorize('int64(int64, int64)')
def _kronecker_delta(a, b):
    if a == b:
        return 1
    else:
        return 0


@vectorize('float64(int64, int64, int64, int64)')
def _moment0(j, l, m, s):
    return _kronecker_delta(j, l)


@vectorize('float64(int64, int64, int64, int64)')
def _moment1(j, l, m, s):
    return np.sqrt((2*l+1)/(2*j+1)) * clebsch_gordan(j, m, 1, 0, l, m) * clebsch_gordan(j, -s, 1, 0, l, -s)


@vectorize('float64(int64, int64, int64, int64)')
def _moment2(j, l, m, s):
    return (_kronecker_delta(j, l)/3.0
            + 2*np.sqrt((2*l+1)/(2*j+1)) * clebsch_gordan(j, m, 2, 0, l, m) * clebsch_gordan(j, -s, 2, 0, l, -s) / 3.0)


def _operator_matrix(m, aomega, s, ell_max):
    # Initialize with empty matrix
    matrix = np.zeros((ell_max+1, ell_max+1), dtype=complex)

    # Set the bands individually
    # 2 above the diagonal:
    ell_range = np.arange(0, ell_max+1-2)
    matrix.flat[2:((ell_max + 1) * (ell_max + 1 - 2)):ell_max + 1 + 1] = (
        aomega**2 * _moment2(ell_range+2, ell_range, m, s)
    )
    # 1 above the diagonal:
    ell_range = np.arange(0, ell_max+1-1)
    matrix.flat[1:((ell_max + 1) * (ell_max + 1 - 1)):ell_max + 1 + 1] = (
        aomega**2 * _moment2(ell_range+1, ell_range, m, s)
        - 2 * aomega * _moment1(ell_range+1, ell_range, m, s)
    )
    # On the diagonal:
    ell_range = np.arange(0, ell_max+1)
    matrix.flat[0:((ell_max + 1) * (ell_max + 1)):ell_max + 1 + 1] = (
        aomega**2 * _moment2(ell_range, ell_range, m, s)
        - 2 * aomega * _moment1(ell_range, ell_range, m, s)
        - ell_range*(ell_range+1)
    )
    # 1 below the diagonal:
    ell_range = np.arange(1, ell_max+1)
    matrix.flat[(ell_max + 1):((ell_max + 1) * (ell_max + 1)):ell_max + 1 + 1] = (
        aomega**2 * _moment2(ell_range-1, ell_range, m, s)
        - 2 * aomega * _moment1(ell_range-1, ell_range, m, s)
    )
    # 2 below the diagonal:
    ell_range = np.arange(2, ell_max+1)
    matrix.flat[2 * (ell_max + 1):((ell_max + 1) * (ell_max + 1)):ell_max + 1 + 1] = (
        aomega**2 * _moment2(ell_range-2, ell_range, m, s)
    )

    return matrix












# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

"""Module for computing Wigner's 3-j symbols

This module uses code from sympy, originally written by Jens Rasch.

"""

from __future__ import print_function, division, absolute_import

from math import sqrt
from .. import factorials
from quaternion.numba_wrapper import njit, xrange


@njit('f8(i8,i8,i8,i8,i8,i8)')
def Wigner3j(j_1, j_2, j_3, m_1, m_2, m_3):
    """Calculate the Wigner 3j symbol `Wigner3j(j_1,j_2,j_3,m_1,m_2,m_3)`

    This function is copied with minor modification from
    sympy.physics.Wigner, as written by Jens Rasch.

    The inputs must be integers.  (Half integer arguments are
    sacrificed so that we can use numba.)  Nonzero return quantities
    only occur when the `j`s obey the triangle inequality (any two
    must add up to be as big as or bigger than the third).

    Examples
    ========

    >>> from spherical_functions import Wigner3j
    >>> Wigner3j(2, 6, 4, 0, 0, 0)
    0.186989398002
    >>> Wigner3j(2, 6, 4, 0, 0, 1)
    0

    """
    if (m_1 + m_2 + m_3 != 0):
        return 0
    if ( (abs(m_1) > j_1) or (abs(m_2) > j_2) or (abs(m_3) > j_3) ):
        return 0
    prefid = (1 if (j_1 - j_2 - m_3) % 2 == 0 else -1)
    m_3 = -m_3
    a1 = j_1 + j_2 - j_3
    if (a1 < 0):
        return 0
    a2 = j_1 - j_2 + j_3;
    if (a2 < 0):
        return 0
    a3 = -j_1 + j_2 + j_3;
    if (a3 < 0):
        return 0

    argsqrt = ( factorials[j_1 + j_2 - j_3] *
                factorials[j_1 - j_2 + j_3] *
                factorials[-j_1 + j_2 + j_3] *
                factorials[j_1 - m_1] *
                factorials[j_1 + m_1] *
                factorials[j_2 - m_2] *
                factorials[j_2 + m_2] *
                factorials[j_3 - m_3] *
                factorials[j_3 + m_3] ) / factorials[j_1 + j_2 + j_3 + 1]

    ressqrt = sqrt(argsqrt)

    imin = max(-j_3 + j_1 + m_2, max(-j_3 + j_2 - m_1, 0))
    imax = min(j_2 + m_2, min(j_1 - m_1, j_1 + j_2 - j_3))
    sumres = 0.0;
    for ii in xrange(imin, imax + 1):
        den = ( factorials[ii] *
                factorials[ii + j_3 - j_1 - m_2] *
                factorials[j_2 + m_2 - ii] *
                factorials[j_1 - ii - m_1] *
                factorials[ii + j_3 - j_2 + m_1] *
                factorials[j_1 + j_2 - j_3 - ii] )
        if (ii % 2 == 0):
            sumres = sumres + 1.0 / den
        else:
            sumres = sumres - 1.0 / den

    return ressqrt * sumres * prefid


@njit('f8(i8,i8,i8,i8,i8,i8)')
def clebsch_gordan(j_1, m_1, j_2, m_2, j_3, m_3):
    """Calculate the Clebsch-Gordan coefficient <j1 m1 j2 m2 | j3 m3>"""
    return (-1.)**(j_1-j_2+m_3) * sqrt(2*j_3+1) * Wigner3j(j_1, j_2, j_3, m_1, m_2, -m_3)



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Modes class.  In
### particular, they assume that the first argument, `self` is an instance of Modes.  They should
### probably not be used outside of that class.


def index(self, ell, m):
    """Return index of (ell, m) mode in data

    Note that the modes are always stored in the last axis of the data, so if your Modes object
    `modes` has (or even just might have) more than one dimension, you might extract a
    particular mode weight with something like this:

        i = modes.index(ell, m)
        ell_m_modes = modes[..., i]

    The ellipsis just represents all other dimensions (even if there are none).

    """
    from .. import LM_index
    if ell < abs(self.s) or ell < abs(m):
        raise ValueError(f"Bad index (ell, m)=({ell}, {m}) for spin weight s={self.s}")
    if ell < self.ell_min or ell > self.ell_max:
        raise ValueError(f"Requested ell index {ell} outside bounds of this data ({self.ell_min, self.ell_max})")
    return LM_index(ell, m, self.ell_min)


def truncate_ell(self, new_ell_max):
    """Slice array so that new ell max is the given value"""
    if new_ell_max >= self.ell_max:
        return self
    truncated = self[..., :self.index(new_ell_max, new_ell_max)+1]
    truncated._metadata['ell_max'] = new_ell_max
    return truncated


def grid(self, n_theta=None, n_phi=None, **kwargs):
    """Return values of function on an equi-angular grid

    This method uses `spinsfast` to convert mode weights of spin-weighted function to values on a
    grid.  The grid has `n_theta` evenly spaced points along the usual polar (colatitude) angle
    theta, and `n_phi` evenly spaced points along the usual azimuthal angle phi.  This grid
    corresponds to the one produced by `spherical_functions.theta_phi`; see that function for
    specifics.

    The output array has one more dimension than this object; rather than the last axis giving the
    mode weights, the last two axes give the values on the two-dimensional grid.

    Parameters
    ==========
    n_theta: None or int [defaults to None]
        Number of points to use in theta direction.  None is equivalent to 2*self.ell_max+1, which
        is the minimum number that can capture behavior up to and including ell_max.  If you need to
        multiply the result with some `other` spin-weighted function, you should use an n_theta
        value of 2 * (self.ell_max + other.ell_max) + 1 to avoid aliasing.
    n_phi: None or int [defaults to None]
        Number of points to use in the phi direction.  Here, None is equivalent to n_phi=n_theta,
        after calculation of the default value for n_theta.  Note that the same comments apply about
        avoiding aliasing.
    **kwargs: any types
        Additional keyword arguments are passed through to the Grid constructor on output

    """
    import copy
    import numpy as np
    import spinsfast
    from .. import Grid
    n_theta = n_theta or 2*self.ell_max+1
    n_phi = n_phi or n_theta
    metadata = copy.copy(self._metadata)
    metadata.pop('ell_max', None)
    metadata.update(**kwargs)
    return Grid(spinsfast.salm2map(self.view(np.ndarray), self.s, self.ell_max, n_theta, n_phi), **metadata)


def evaluate(self, rotors, **kwargs):
    """Return values of function on input rotors"""
    import numpy as np
    import spherical_functions as sf
    SWSHs = sf.SWSH_grid(rotors, self.spin_weight, self.ell_max)
    return np.tensordot(
        self.view(np.ndarray),
        SWSHs[..., sf.LM_index(self.ell_min, -self.ell_min, 0):sf.LM_index(self.ell_max, self.ell_max, 0)+1],
        axes=([-1], [-1])
    )


def _check_broadcasting(self, array, reverse=False):
    """Test whether or not the given array can broadcast against this object"""
    import numpy as np

    if isinstance(array, type(self)):
        try:
            if reverse:
                np.broadcast(array, self)
            else:
                np.broadcast(self, array)
        except ValueError:
            return False
        else:
            return True
    else:
        if np.ndim(array) > np.ndim(self)-1:
            raise ValueError(f"Cannot broadcast array of {np.ndim(array)} dimensions against {type(self).__name__} "
                             f"object of fewer ({np.ndim(self)-1}) non-mode dimensions.\n"
                             "This is to ensure that scalars do not operate on individual "
                             "mode weights; they must operate on all simultaneously.\n"
                             "If that is the case and you still want to broadcast, add more "
                             "dimensions before this object's first dimension.")
        try:
            if reverse:
                np.broadcast(array, self[..., 0])
            else:
                np.broadcast(self[..., 0], array)
        except ValueError:
            return False
        else:
            return True



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

import copy
import math
import numpy as np
from .. import LM_total_size, Wigner3j, LM_index, LM_deduce_ell_max
from ..multiplication import _multiplication_helper


class Modes(np.ndarray):
    """Object to store SWHS modes

    This class subclasses numpy's ndarray object, so that it should act like a numpy array in many
    respects, even with functions like np.zeros_like.

    NOTE: The functions `np.copy(modes)` and `np.array(modes, copy=True)` return `ndarray` objects;
    they lose information about the SWSH attributes, and have different ufuncs.  If you wish to keep
    this information, use `modes.copy()`.  Also note that pickling works as expected, as do
    copy.copy and copy.deepcopy.

    The number of dimensions is arbitrary as long as it is at least 1, but the modes must be stored
    in the last axis.  For example, a SWSH function of time may be stored as a 2-d array where the
    first axis represents different times, and the second axis represents the mode weights at each
    instant of time.

    This class also does three important things that are unlike numpy arrays:

    1) It tracks the spin weight and minimum and maximum ell values stored in this data.

    2) It provides additional convenience methods, like `index` to find the index of a particular
       (ell, m) mode; a `grid` method to convert modes to the SWSH values on a grid over the sphere
       (while correctly handling additional dimensions); and the various derivative operators,
       including $\eth$ and $\bar{\eth}$.

    3) It overrides most of numpy's "universal functions" (ufuncs) to work appropriately for
       spin-weighted functions.  Specifically, these ufuncs are interpreted as acting on the
       spin-weighted function itself, rather than just the mode weights.  Most importantly, we have
       these overriden methods:

       a) Conjugating a Modes object will result in a new Modes object that represents the
          conjugated spin-weighted function, rather than simply conjugating the mode-weights of the
          function.

       b) Multiplying two Modes objects will result in a new Modes object that represents the
          pointwise product of the functions themselves (and will correctly have spin weight given
          by the sum of the spin weights of the first two functions), rather than the product of the
          mode weights.  Division is only permitted when the divisor is a constant.

       c) Addition (and subtraction) is permitted for functions of the same spin weight, but it does
          not make sense to add (or subtract) functions with different spin weights, so any attempt
          to do so raises a ValueError.  [Note that adding a constant is equivalent to adding a
          function of spin weight zero, and is treated in the same way.]

       d) The "absolute" ufunc does not return the absolute value of each mode weight (which is
          almost certainly meaningless); it returns the L2 norm of spin-weighted function over the
          sphere -- which happens to equal the sum of the squares of the absolute values of the mode
          weights.

       Numerous other ufuncs -- such as log, exp, trigonometric ufuncs, bit-twiddling ufuncs, and so
       on -- are disabled because they don't make sense when applied to functions.

    It is possible to treat the underlying data of a Modes object `modes` as an ordinary numpy array
    by taking `modes.view(np.ndarray)`.  However, it is hoped that this class already performs all
    reasonable operations.  If you find a missing feature that requires you to resort to this,
    please feel free to open an issue in this project's github page to discuss it.

    Also, be aware that ndarrays also have various built-in methods that cannot be overridden very
    easily, such as max, copysign, etc.  If you try to use -- even indirectly -- those functions
    that don't have any clear interpretation for spin-weighted functions, things will likely break.


    Constructor parameters
    ======================
    input_array: array_like
        This may be a numpy ndarray, or any subclass.  It must be able to be viewed as complex.  If
        it has a `_metadata` attribute, that field will be copied to the new array; if the following
        parameters are not passed to the constructor, they will be searched for in the metadata.
    spin_weight: int [optional if present in `input_array._metadata`]
        The spin weight of the function that this Modes object describes.  This must be specified
        somehow, whether via a `_metadata` attribute of the input array, or as a keyword argument,
        or as the second positional argument (where the latter will override the former values).
    ell_min: int [defaults to 0]
        The smallest ell value present in the *input* data.  Note that the data will be stored in
        this object with all modes present, starting from ell=0.  This parameter just determines how
        the input data will be copied into the internal representation.
    ell_max: int [optional if ell_min is not passed as positional argument]
        The largest ell value present in the input data.  If this is not passed explicitly, it will
        be inferred from the size of the data.
    multiplication_truncator: None or callable [optional]
        Function to be used by default when multiplying Modes objects together.  See the
        documentation for spherical_functions.Modes.multiply for more details.  The default behavior
        with `sum` is the most correct one -- keeping all ell values that result -- but also the
        most wasteful, and very likely to be overkill.  The user may prefer to use `max`, which will
        just return a product with ell_max equal to the larger of the two input ell_max values.

    """

    # https://numpy.org/doc/1.18/user/basics.subclassing.html
    def __new__(cls, input_array, *args, **kwargs):
        if len(args) == 2 or len(args) > 3:
            raise ValueError("Only one, two, or four positional arguments may be passed")
        if len(args) >= 1:
            kwargs['spin_weight'] = args[0]
        if len(args) == 3:
            kwargs['ell_min'] = args[1]
            kwargs['ell_max'] = args[2]
        ell_min = kwargs.pop('ell_min', 0)
        metadata = copy.copy(getattr(input_array, '_metadata', {}))
        metadata.update(**kwargs)
        input_array = np.asanyarray(input_array).view(complex)
        spin_weight = metadata.get('spin_weight', None)
        if spin_weight is None:
            raise ValueError("Spin weight must be specified")
        ell_max = metadata.get('ell_max', None)
        if ell_max is None:
            if np.ndim(input_array) == 0:
                ell_max = 0
            else:
                ell_max = LM_deduce_ell_max(input_array.shape[-1], ell_min)
            metadata['ell_max'] = ell_max
        if input_array.shape[-1] != LM_total_size(ell_min, ell_max):
            raise ValueError(f"Input array has shape {input_array.shape} when viewed as a complex array.\n            "
                             +f"Its last dimension should have size {LM_total_size(ell_min, ell_max)}, "
                             +f"for consistency with the input ell_min ({ell_min}) and ell_max ({ell_max}).")
        if ell_min == 0:
            obj = input_array.view(cls)
        else:
            insertion_indices = [0,]*LM_total_size(0, ell_min-1)
            obj = np.insert(input_array, insertion_indices, 0.0, axis=-1).view(cls)
        obj[..., :LM_total_size(0, abs(spin_weight)-1)] = 0.0
        obj._metadata = metadata
        return obj

    # https://numpy.org/doc/1.18/user/basics.subclassing.html
    def __array_finalize__(self, obj):
        if obj is None: return
        self._metadata = copy.copy(getattr(obj, '_metadata', {}))
        if not 'spin_weight' in self._metadata:
            self._metadata['spin_weight'] = None
        if not 'ell_max' in self._metadata:
            self._metadata['ell_max'] = None

    # For pickling
    def __reduce__(self):
        state = super(Modes, self).__reduce__()
        new_attributes = state[2] + (self._metadata,)
        return (state[0], state[1], new_attributes)

    # For unpickling
    def __setstate__(self, state):
        self._metadata = copy.deepcopy(state[-1])
        super(Modes, self).__setstate__(state[:-1])

    @property
    def s(self):
        """Spin weight of this Modes object"""
        return self._metadata['spin_weight']

    spin_weight = s

    @property
    def ell_min(self):
        """Smallest ell value stored in data [need not equal abs(s)]"""
        return 0

    @property
    def ell_max(self):
        """Largest ell value stored in data"""
        return self._metadata['ell_max']

    @property
    def n_modes(self):
        """Number of elements along the last axis"""
        return self.shape[-1]

    from .algebra import (
        conjugate, bar, _real_func, real, _imag_func, imag, norm,
        add, subtract, multiply, divide
    )

    conj = conjugate

    from .derivatives import (
        Lsquared, Lz, Lplus, Lminus,
        Rsquared, Rz, Rplus, Rminus,
        eth, ethbar
    )

    from .utilities import (
        index, truncate_ell, grid, evaluate, _check_broadcasting
    )

    from .ufuncs import __array_ufunc__



# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

import math
import cmath

import numpy as np
import quaternion
from quaternion.numba_wrapper import njit, jit, int64, xrange

from .. import (Wigner_coefficient as coeff, epsilon, LM_range)


def SWSH(R, s, indices):
    """Spin-weighted spherical harmonic calculation from rotor

    Note that this function is more general than standard Ylm and sYlm functions of angles because it uses quaternion
    rotors instead of angle, and is slightly faster as a result.

    This function can be called in either of two ways:
      1) With an array of quaternions, and a single (ell,m) pair, or
      2) with a single quaternion, and an array of (ell,m) values

    Parameters
    ----------
    R : unit quaternion or array of unit quaternions
        Rotor on which to evaluate the SWSH function.
    s : int
        Spin weight of the field to evaluate
    indices : 2-d array of int or pair of ints
        Array of (ell,m) values to evaluate

    Returns
    -------
    array of complex
        The shape of this array is `indices.shape[0]`, and contains the values of the SWSH for the (ell,m) values
        specified in `indices`.

    """
    indices = np.asarray(indices)
    if indices.size > 2 or not isinstance(R, np.ndarray):
        values = np.empty((indices.shape[0],), dtype=complex)
        _SWSH(R.a, R.b, s, indices, values)
    else:
        values = np.empty((R.size,), dtype=complex)
        _SWSHs(quaternion.as_float_array(R.flatten()), s, indices[0], indices[1], values)
        values = values.reshape(R.shape)
    return values


def SWSH_grid(R_grid, s, ell_max):
    """Spin-weighted spherical harmonic calculation from rotors representing a grid

    This function is similar to the `SWSH` function, but assumes that the input is an array of rotors, representing
    various points and/or orientations for which to evaluate the SWSH values.  It simply iterates over each of the
    quaternions, and evaluates the separately for each rotor.

    Note that this function takes `ell_max` as its argument, rather than a set of indices.  The output of this
    function is simply the entire range of indices.

    Also note that his function is more general than standard Ylm and sYlm functions of angles because it uses
    quaternion rotors instead of angle, and is slightly faster as a result.

    Parameters
    ----------
    R_grid : array of unit quaternions
        Rotors on which to evaluate the SWSH function.  Note that, for speed, these are assumed to be normalized.
    s : int
        Spin weight of the field to evaluate
    ell_max : int
        Largest ell value in output arrays.  Note that this should probably be `ell_max >= abs(s)`, but the output
        array will contain values corresponding to `ell < abs(s)`.  Those values will be 0.0, but must be present for
        compatibility with `spinsfast`.

    Returns
    -------
    array of complex
        The shape of this array is `R_grid.shape`, with an extra dimension of length N_lm appended.  That extra
        dimension corresponds to the various (ell,m) values in standard order (e.g., as given by LM_range),
        starting from ell=0 for compatibility with spinsfast.

    """
    indices = LM_range(0, ell_max)
    values = np.zeros(R_grid.shape + (indices.shape[0],), dtype=complex)
    it = np.nditer(R_grid, flags=['multi_index'])
    while not it.finished:
        R = it[0][()]
        _SWSH(R.a, R.b, s, indices, values[it.multi_index])
        it.iternext()
    return values


@njit('void(complex128, complex128, int64, int64[:,:], complex128[:])')
def _SWSH(Ra, Rb, s, indices, values):
    """Compute spin-weighted spherical harmonics from rotor components

    This is the core function that does all the work in the
    computation, but it is strict about its inputs, and does not check
    them for validity -- though numba provides some degree of safety.

    _SWSH(Ra, Rb, s, indices, values)

    Parameters
    ----------
    Ra : complex
        Component `a` of the rotor
    Rb : complex
        Component `b` of the rotor
    s : int
        Spin weight of the field to evaluate
    indices : 2-d array of int
        Array of (ell,m) values to evaluate
    values : 1-d array of complex
        Output array to contain values.  Length must equal first dimension of `indices`.  Needed because numba cannot
        create arrays at the moment.

    Returns
    -------
    void
        The input/output array `values` is modified in place.

    """
    N = indices.shape[0]

    # These constants are the recurring quantities in the computation
    # of the matrix values, so we calculate them here just once

    ra, phia = cmath.polar(Ra)
    rb, phib = cmath.polar(Rb)

    if ra <= epsilon:
        for i in xrange(N):
            ell, m = indices[i, 0:2]
            if (m != s or abs(m) > ell or abs(s) > ell):
                values[i] = 0.0j
            else:
                if (ell) % 2 == 0:
                    values[i] = math.sqrt((2 * ell + 1) / (4 * np.pi)) * Rb ** (-2 * s)
                else:
                    values[i] = -math.sqrt((2 * ell + 1) / (4 * np.pi)) * Rb ** (-2 * s)

    elif rb <= epsilon:
        for i in xrange(N):
            ell, m = indices[i, 0:2]
            if (m != -s or abs(m) > ell or abs(s) > ell):
                values[i] = 0.0j
            else:
                if (s) % 2 == 0:
                    values[i] = math.sqrt((2 * ell + 1) / (4 * np.pi)) * Ra ** (-2 * s)
                else:
                    values[i] = -math.sqrt((2 * ell + 1) / (4 * np.pi)) * Ra ** (-2 * s)

    elif ra < rb:
        # We have to have these two versions (both this ra<rb branch,
        # and ra>=rb below) to avoid overflows and underflows
        absRRatioSquared = -ra * ra / (rb * rb)
        for i in xrange(N):
            ell, m = indices[i, 0:2]
            if (abs(m) > ell or abs(s) > ell):
                values[i] = 0.0j
            else:
                rhoMin = max(0, -m + s)
                # Protect against overflow by decomposing Ra,Rb as
                # abs,angle components and pulling out the factor of
                # absRRatioSquared**rhoMin.  Here, ra might be quite
                # small, in which case ra**(-s+m) could be enormous
                # when the exponent (-s+m) is very negative; adding
                # 2*rhoMin to the exponent ensures that it is always
                # positive, which protects from overflow.  Meanwhile,
                # underflow just goes to zero, which is fine since
                # nothing else should be very large.
                Prefactor = cmath.rect(
                    coeff(ell, -m, -s) * rb ** (2 * ell + s - m - 2 * rhoMin) * ra ** (-s + m + 2 * rhoMin),
                    phib * (-s - m) + phia * (-s + m))
                if (Prefactor == 0.0j):
                    values[i] = 0.0j
                else:
                    if ((ell + rhoMin) % 2 != 0):
                        Prefactor *= -1
                    rhoMax = min(ell - m, ell + s)
                    N1 = ell - m + 1
                    N2 = ell + s + 1
                    M = -s + m
                    Sum = 1.0
                    for rho in xrange(rhoMax, rhoMin, -1):
                        Sum *= absRRatioSquared * ((N1 - rho) * (N2 - rho)) / (rho * (M + rho))
                        Sum += 1
                    # Sum = 0.0
                    # for rho in xrange(rhoMax, rhoMin-1, -1):
                    # Sum = (  binomial_coefficient(ell-m,rho) * binomial_coefficient(ell+m, ell-rho+s)
                    # + Sum * absRRatioSquared )
                    values[i] = math.sqrt((2 * ell + 1) / (4 * np.pi)) * Prefactor * Sum

    else:  # ra >= rb
        # We have to have these two versions (both this ra>=rb branch,
        # and ra<rb above) to avoid overflows and underflows
        absRRatioSquared = -rb * rb / (ra * ra)
        for i in xrange(N):
            ell, m = indices[i, 0:2]
            if (abs(m) > ell or abs(s) > ell):
                values[i] = 0.0j
            else:
                rhoMin = max(0, m + s)
                # Protect against overflow by decomposing Ra,Rb as
                # abs,angle components and pulling out the factor of
                # absRRatioSquared**rhoMin.  Here, rb might be quite
                # small, in which case rb**(-s-m) could be enormous
                # when the exponent (-s-m) is very negative; adding
                # 2*rhoMin to the exponent ensures that it is always
                # positive, which protects from overflow.  Meanwhile,
                # underflow just goes to zero, which is fine since
                # nothing else should be very large.
                Prefactor = cmath.rect(
                    coeff(ell, m, -s) * ra ** (2 * ell + s + m - 2 * rhoMin) * rb ** (-s - m + 2 * rhoMin),
                    phia * (-s + m) + phib * (-s - m))
                if (Prefactor == 0.0j):
                    values[i] = 0.0j
                else:
                    if ((rhoMin + s) % 2 != 0):
                        Prefactor *= -1
                    rhoMax = min(ell + m, ell + s)
                    N1 = ell + m + 1
                    N2 = ell + s + 1
                    M = -s - m
                    Sum = 1.0
                    for rho in xrange(rhoMax, rhoMin, -1):
                        Sum *= absRRatioSquared * ((N1 - rho) * (N2 - rho)) / (rho * (M + rho))
                        Sum += 1
                    # Sum = 0.0
                    # for rho in xrange(rhoMax, rhoMin-1, -1):
                    # Sum = (  binomial_coefficient(ell+m,rho) * binomial_coefficient(ell-m, ell-rho+s)
                    # + Sum * absRRatioSquared )
                    values[i] = math.sqrt((2 * ell + 1) / (4 * np.pi)) * Prefactor * Sum


@njit('void(float64[:,:], int64, int64, int64, complex128[:])')
def _SWSHs(Rs, s, ell, m, values):
    """Compute spin-weighted spherical harmonics from rotor components

    This is the core function that does all the work in the
    computation, but it is strict about its inputs, and does not check
    them for validity -- though numba provides some degree of safety.

    _SWSHs(Rs, s, ell, m, values)

    Parameters
    ----------
    Rs : 2-d array of float
        Components of the rotors, with the 0 index iterating over rotor, and the 1 index iterating over component.
    s : int
        Spin weight of the field to evaluate
    ell : int
    m : int
        Values of (ell,m) to output
    values : 1-d array of complex
        Output array to contain values.  Length must equal 0 dimension of `Rs`.  Needed because numba cannot create
        arrays at the moment.

    Returns
    -------
    void
        The input/output array `values` is modified in place.

    """
    N = Rs.shape[0]

    if (abs(m) > ell or abs(s) > ell):
        for i in xrange(N):
            values[i] = 0.0j

    else:

        constant = math.sqrt((2 * ell + 1) / (4 * np.pi))
        ell_even = (ell % 2 == 0)
        s_even = (s % 2 == 0)
        rhoMin_a = max(0, m + s)
        rhoMax_a = min(ell + m, ell + s)
        coefficient_a = coeff(ell, m, -s)
        if ((rhoMin_a + s) % 2 != 0):
            coefficient_a *= -1
        N1_a = ell + m + 1
        N2_a = ell + s + 1
        M_a = -s - m
        rhoMin_b = max(0, -m + s)
        rhoMax_b = min(ell - m, ell + s)
        coefficient_b = coeff(ell, -m, -s)
        if ((ell + rhoMin_b) % 2 != 0):
            coefficient_b *= -1
        N1_b = ell - m + 1
        N2_b = ell + s + 1
        M_b = -s + m

        for i in xrange(N):
            Ra = complex(Rs[i, 0], Rs[i, 3])
            ra, phia = cmath.polar(Ra)

            Rb = complex(Rs[i, 2], Rs[i, 1])
            rb, phib = cmath.polar(Rb)

            if ra <= epsilon:
                if m != s:
                    values[i] = 0.0j
                elif ell_even:
                    values[i] = constant * Rb ** (-2 * s)
                else:
                    values[i] = -constant * Rb ** (-2 * s)

            elif rb <= epsilon:
                if m != -s:
                    values[i] = 0.0j
                elif s_even:
                    values[i] = constant * Ra ** (-2 * s)
                else:
                    values[i] = -constant * Ra ** (-2 * s)

            elif ra < rb:
                if (coefficient_b == 0.0j):
                    values[i] = 0.0j
                else:
                    absRRatioSquared = -ra * ra / (rb * rb)
                    Prefactor = cmath.rect(
                        coefficient_b * rb ** (2 * ell + s - m - 2 * rhoMin_b) * ra ** (-s + m + 2 * rhoMin_b),
                        phib * (-s - m) + phia * (-s + m))
                    Sum = 1.0
                    for rho in xrange(rhoMax_b, rhoMin_b, -1):
                        Sum *= absRRatioSquared * ((N1_b - rho) * (N2_b - rho)) / (rho * (M_b + rho))
                        Sum += 1
                    values[i] = constant * Prefactor * Sum

            else:  # ra >= rb
                if (coefficient_a == 0.0j):
                    values[i] = 0.0j
                else:
                    absRRatioSquared = -rb * rb / (ra * ra)
                    Prefactor = cmath.rect(
                        coefficient_a * ra ** (2 * ell + s + m - 2 * rhoMin_a) * rb ** (-s - m + 2 * rhoMin_a),
                        phia * (-s + m) + phib * (-s - m))
                    Sum = 1.0
                    for rho in xrange(rhoMax_a, rhoMin_a, -1):
                        Sum *= absRRatioSquared * ((N1_a - rho) * (N2_a - rho)) / (rho * (M_a + rho))
                        Sum += 1
                    values[i] = constant * Prefactor * Sum



import numpy as np
from functools import lru_cache
from scipy.special import factorial
from numba import njit
import spherical_functions as sf

"""Algorithm for computing H, as given by arxiv:1403.7698:

Symmetry relations:
H^{m', m}_n(\beta) = H^{m, m'}_n(\beta)
H^{m', m}_n(\beta) = H^{-m', -m}_n(\beta)
H^{m', m}_n(\beta) = (-1)^{n+m+m'} H^{-m', m}_n(\pi - \beta)
H^{m', m}_n(\beta) = (-1)^{m+m'} H^{m', m}_n(-\beta)

1: If n=0 set H_{0}^{0,0}=1.

2: Compute values H^{0,m}_{n}(β)for m=0,...,n and H^{0,m}_{n+1}(β) for m=0,...,n+1 using Eq. (32):
     H^{0,m}_{n}(β) = (-1)^m \sqrt{(n-|m|)!/(n+|m|)!} P^{|m|}_{n}(cos β)

3: Use relation (41) to compute H^{1,m}_{n}(β) for m=1,...,n.  Using symmetry and shift of the
   indices this relation can be written as
     b^{0}_{n+1} H^{1,m}_{n} = \frac{b^{−m−1}_{n+1} (1−cosβ)}{2} H^{0,m+1}_{n+1}
                             − \frac{b^{m−1}_{n+1}  (1+cosβ)}{2} H^{0,m−1}_{n+1}
                             − a^{m}_{n} sinβ H^{0,m}_{n+1}

4: Recursively compute H^{m′+1,m}_{n}(β) for m′=1,...,n−1, m=m′,...,n using relation (50) resolved
   with respect to H^{m′+1,m}_{n}:
     d^{m′}_{n} H^{m′+1,m}_{n} = d^{m′−1}_{n} H^{m′−1,m}_{n}
                               − d^{m−1}_{n} H^{m′,m−1}_{n}
                               + d^{m}_{n} H^{m′,m+1}_{n}
   (where the last term drops out for m=n).

5. Recursively compute H^{m′−1,m}_{n}(β) for m′=−1,...,−n+1, m=−m′,...,n using relation (50)
   resolved with respect to H^{m′−1,m}_{n}:
     d^{m′−1}_{n} H^{m′−1,m}_{n} = d^{m′}_{n} H^{m′+1,m}_{n}
                                 + d^{m−1}_{n} H^{m′,m−1}_{n}
                                 − d^{m}_{n} H^{m′,m+1}_{n}
   (where the last term drops out for m=n).

6: Apply the first and the second symmetry relations above to obtain all other values H^{m′,m}_{n}
   outside the computational triangle m=0,...,n, m′=−m,...,m.

"""

"""Associated Legendre function recursion

All references are to "A unified approach to the Clenshaw summation and the recursive computation of
very high degree and order normalised associated Legendre functions" by Holmes and Featherstone
(2002).  The text suggests that the best approach -- both in terms of speed and accuracy -- is the
"modified forward row method" described in Sec. 2.5.

Define k=1 for m=0 and k=2 for m>0, j=2 for m=0 and j=1 for m>0, and

\hat{P}_{n, m}(\theta) = \sqrt{ \frac{k(2n+1)(n-m)!} {(n+m)!} } P_{n,m}(\theta) / \sin^m(\theta)

Now, define the coefficients

g_{n,m} = \frac{2(m+1)} {\sqrt{(n-m)(n+m+1)}},

and

h_{n,m} = \sqrt{ \frac{(n+m+2)(n-m-1)} {(n-m)(n+m+1)} }.

Noting that h_{n,n-1} is always 0, we formulate the recursion as

\hat{P}_{1,1} = \sqrt{3},

\hat{P}_{n,n} = \sqrt{\frac{2n+1} {2n}} \hat{P}_{n-1,n-1} \text{ for $n>1},

\hat{P}_{n,m} = \frac{1}{\sqrt{j}} \left(g_{n,m} \cos\theta \hat{P}_{n,m+1} - h_{n,m} \sin^2\theta \hat{P}_{n,m+2}\right),

where the last term in the last equation above drops out for $m=n-1$.
"""


@njit
def sign(m):
    if m >= 0:
        return 1
    else:
        return -1


@njit
def nm_index(n, m):
    """Return flat index into arrray of [n, m] pairs.
    
    Assumes array is ordered as [[n, m] for n in range(n_max+1) for m in range(-n, n+1)]

    """
    return m + n * (n + 1)


@njit
def nabsm_index(n, absm):
    """Return flat index into arrray of [n, abs(m)] pairs
    
    Assumes array is ordered as [[n, m] for n in range(n_max+1) for m in range(n+1)]

    """
    return absm + (n * (n + 1)) // 2


@njit
def nmpm_index(n, mp, m):
    """Return flat index into arrray of [n, mp, m]
    
    Assumes array is ordered as [[n, mp, m] for n in range(n_max+1) for mp in range(-n, n+1) for m in range(-n, n+1)]

    """
    return (((4 * n + 6) * n + 6 * mp + 5) * n + 3 * (m + mp)) // 3


@njit
def _step_1(Hnmpm):
    """If n=0 set H_{0}^{0,0}=1."""
    Hnmpm[0, :] = 1.0


@njit
def _step_2(g, h, n_max, Hnmpm, cosβ, sinβ):
    """Compute values H^{0,m}_{n}(β)for m=0,...,n and H^{0,m}_{n+1}(β) for m=0,...,n+1 using Eq. (32):
        H^{0,m}_{n}(β) = (-1)^m \sqrt{(n-|m|)! / (n+|m|)!} P^{|m|}_{n}(cos β)
                       = (-1)^m (sin β)^m \hat{P}^{|m|}_{n}(cos β) / \sqrt{k (2n+1)}

    This function computes the associated Legendre functions directly by recursion as explained by
    Holmes and Featherstone (2002), doi:10.1007/s00190-002-0216-2.  Note that I had to adjust
    certain steps for consistency with the notation assumed by arxiv:1403.7698 -- mostly involving
    factors of (-1)**m.

    NOTE: Though not specified in arxiv:1403.7698, there is not enough information for step 4 unless we
    also use symmetry to set H^{1,0}_{n} here.  Similarly, step 5 needs additional information, which
    depends on setting H^{0, -1}_{n} from its symmetric equivalent H^{0, 1}_{n} in this step.

    """
    sin2β = sinβ**2

    # n = 1
    n0n_index = nmpm_index(1, 0, 1)
    nn_index = nm_index(1, 1)
    Hnmpm[n0n_index, :] = np.sqrt(3)  # Un-normalized
    Hnmpm[n0n_index-1, :] = (g[nn_index-1] * cosβ) / np.sqrt(2)  # Normalized
    # n = 2, ..., n_max+1
    for n in range(2, n_max+2):
        n0n_index = nmpm_index(n, 0, n)
        nm10nm1_index = nmpm_index(n-1, 0, n-1)
        nn_index = nm_index(n, n)
        const = np.sqrt((2*n+1) / (2*n))
        gi = g[nn_index-1]
        for j in range(Hnmpm.shape[1]):
            # m = n
            Hnmpm[n0n_index, j] = const * Hnmpm[nm10nm1_index, j]
            # m = n-1
            Hnmpm[n0n_index-1, j] = gi * cosβ[j] * Hnmpm[n0n_index, j]
        # m = n-2, ..., 1
        for i in range(2, n):
            gi = g[nn_index-i]
            hi = h[nn_index-i]
            for j in range(Hnmpm.shape[1]):
                Hnmpm[n0n_index-i, j] = gi * cosβ[j] * Hnmpm[n0n_index-i+1, j] - hi * sin2β[j] * Hnmpm[n0n_index-i+2, j]
        # m = 0, with normalization
        const = np.sqrt(2*(2*n+1))
        gi = g[nn_index-n]
        hi = h[nn_index-n]
        for j in range(Hnmpm.shape[1]):
            Hnmpm[n0n_index-n, j] = (gi * cosβ[j] * Hnmpm[n0n_index-n+1, j] - hi * sin2β[j] * Hnmpm[n0n_index-n+2, j]) / const
        # Now, loop back through, correcting the normalization for this row, except for n=n element
        prefactor = np.full_like(sinβ, 1/const)
        for i in range(1, n):
            prefactor *= sinβ
            Hnmpm[n0n_index-n+i, :] *= prefactor
        # Supply extra edge cases as noted in docstring
        Hnmpm[nmpm_index(n, 1, 0), :] = Hnmpm[nmpm_index(n, 0, 1)]
        Hnmpm[nmpm_index(n, 0, -1), :] = Hnmpm[nmpm_index(n, 0, 1)]
    # Correct normalization of m=n elements
    prefactor = np.ones_like(sinβ)
    for n in range(1, n_max+2):
        prefactor *= sinβ
        Hnmpm[nmpm_index(n, 0, n), :] *= prefactor / np.sqrt(2*(2*n+1))
    # Supply extra edge cases as noted in docstring
    Hnmpm[nmpm_index(1, 1, 0), :] = Hnmpm[nmpm_index(1, 0, 1)]
    Hnmpm[nmpm_index(1, 0, -1), :] = Hnmpm[nmpm_index(1, 0, 1)]


@njit
def _step_3(a, b, n_max, Hnmpm, cosβ, sinβ):
    """Use relation (41) to compute H^{1,m}_{n}(β) for m=1,...,n.  Using symmetry and shift of the
    indices this relation can be written as
        b^{0}_{n+1} H^{1, m}_{n} =   \frac{b^{−m−1}_{n+1} (1−cosβ)}{2} H^{0, m+1}_{n+1}
                                   − \frac{b^{ m−1}_{n+1} (1+cosβ)}{2} H^{0, m−1}_{n+1}
                                   − a^{m}_{n} sinβ H^{0, m}_{n+1}
    """
    # a = self.a.reshape(self.a.shape + (1,)*(Hnmpm.ndim-1))
    for n in range(1, n_max+1):
        # m = 1, ..., n
        i1 = nmpm_index(n, 1, 1)
        i2 = nmpm_index(n+1, 0, 2)
        i3 = nmpm_index(n+1, 0, 0)
        i4 = nmpm_index(n+1, 0, 1)
        i5 = nm_index(n+1, 0)
        i6 = nm_index(n+1, -2)
        i7 = nm_index(n+1, 0)
        i8 = nabsm_index(n, 1)
        b5 = b[i5]
        for i in range(n):
            b6 = b[-i+i6]
            b7 = b[i+i7]
            a8 = a[i+i8]
            for j in range(Hnmpm.shape[1]):
                Hnmpm[i+i1, j] = (1 / b5) * (
                    0.5 * (
                          b6 * (1-cosβ[j]) * Hnmpm[i+i2, j]
                        - b7 * (1+cosβ[j]) * Hnmpm[i+i3, j]
                    )
                    - a8 * sinβ[j] * Hnmpm[i+i4, j]
                )


@njit
def _step_4(d, n_max, Hnmpm):
    """Recursively compute H^{m'+1, m}_{n}(β) for m'=1,...,n−1, m=m',...,n using relation (50) resolved
    with respect to H^{m'+1, m}_{n}:
      d^{m'}_{n} H^{m'+1, m}_{n} =   d^{m'−1}_{n} H^{m'−1, m}_{n}
                                   − d^{m−1}_{n} H^{m', m−1}_{n}
                                   + d^{m}_{n} H^{m', m+1}_{n}
    (where the last term drops out for m=n).

    """
    for n in range(2, n_max+1):
        for mp in range(1, n):
            # m = m', ..., n-1
            i1 = nmpm_index(n, mp+1, mp)
            i2 = nmpm_index(n, mp-1, mp)
            i3 = nmpm_index(n, mp, mp-1)
            i4 = nmpm_index(n, mp, mp+1)
            i5 = nm_index(n, mp)
            i6 = nm_index(n, mp-1)
            d5 = d[i5]
            d6 = d[i6]
            for i in range(n-mp):
                d7 = d[i+i6]
                d8 = d[i+i5]
                for j in range(Hnmpm.shape[1]):
                    Hnmpm[i+i1, j] = (1 / d5) * (
                          d6 * Hnmpm[i+i2, j]
                        - d7 * Hnmpm[i+i3, j]
                        + d8 * Hnmpm[i+i4, j]
                    )
            # m = n
            i = n-mp
            for j in range(Hnmpm.shape[1]):
                Hnmpm[i+i1, j] = (1 / d5) * (
                      d6 * Hnmpm[i+i2, j]
                    - d[i+i6] * Hnmpm[i+i3, j]
                )


@njit
def _step_5(d, n_max, Hnmpm):
    """Recursively compute H^{m'−1, m}_{n}(β) for m'=−1,...,−n+1, m=−m',...,n using relation (50)
    resolved with respect to H^{m'−1, m}_{n}:
      d^{m'−1}_{n} H^{m'−1, m}_{n} = d^{m'}_{n} H^{m'+1, m}_{n}
                                     + d^{m−1}_{n} H^{m', m−1}_{n}
                                     − d^{m}_{n} H^{m', m+1}_{n}
    (where the last term drops out for m=n).

    NOTE: Although arxiv:1403.7698 specifies the loop over mp to start at -1, I find it necessary to
    start at 0, or there will be missing information.  This also requires setting the (m',m)=(0,-1)
    components before beginning this loop.

    """
    for n in range(0, n_max+1):
        for mp in range(0, -n, -1):
            # m = -m', ..., n-1
            i1 = nmpm_index(n, mp-1, -mp)
            i2 = nmpm_index(n, mp+1, -mp)
            i3 = nmpm_index(n, mp, -mp-1)
            i4 = nmpm_index(n, mp, -mp+1)
            i5 = nm_index(n, mp-1)
            i6 = nm_index(n, mp)
            i7 = nm_index(n, -mp-1)
            i8 = nm_index(n, -mp)
            d5 = d[i5]
            d6 = d[i6]
            for i in range(n+mp):
                d7 = d[i+i7]
                d8 = d[i+i8]
                for j in range(Hnmpm.shape[1]):
                    Hnmpm[i+i1, j] = (1 / d5) * (
                          d6 * Hnmpm[i+i2, j]
                        + d7 * Hnmpm[i+i3, j]
                        - d8 * Hnmpm[i+i4, j]
                    )
            # m = n
            i = n+mp
            for j in range(Hnmpm.shape[1]):
                Hnmpm[i+i1, j] = (1 / d5) * (
                      d6 * Hnmpm[i+i2, j]
                    + d[i+i7] * Hnmpm[i+i3, j]
                )


@njit
def _step_6(n_max, Hnmpm):
    """Apply the symmetry relations below to obtain all other values H^{m',m}_{n}
    outside the computational triangle m=0,...,n, m'=−m,...,m:
      H^{m', m}_n(\beta) = H^{m, m'}_n(\beta)
      H^{m', m}_n(\beta) = H^{-m', -m}_n(\beta)
    """
    for n in range(1, n_max+1):
        for mp in range(1, n+1):
            for m in range(max(-mp-1, -n), mp-1):
                for j in range(Hnmpm.shape[1]):
                    Hnmpm[nmpm_index(n, mp, m), j] = Hnmpm[nmpm_index(n, m, mp), j]
        for mp in range(-n, n+1):
            for m in range(-n, -mp-1):
                for j in range(Hnmpm.shape[1]):
                    Hnmpm[nmpm_index(n, mp, m), j] = Hnmpm[nmpm_index(n, -mp, -m), j]


class HCalculator(object):
    def __init__(self, n_max):
        self.n_max = int(n_max)
        if self.n_max < 0:
            raise ValueError('Nonsensical value for n_max = {0}'.format(self.n_max))
        self.nmpm_total_size = sf.LMpM_total_size(0, self.n_max)
        self.nmpm_total_size_plus1 = sf.LMpM_total_size(0, self.n_max+1)
        n = np.array([n for n in range(self.n_max+2) for m in range(-n, n+1)])
        m = np.array([m for n in range(self.n_max+2) for m in range(-n, n+1)])
        absn = np.array([n for n in range(self.n_max+2) for m in range(n+1)])
        absm = np.array([m for n in range(self.n_max+2) for m in range(n+1)])
        self.sqrt_factorial_ratio = (-1)**absm * np.sqrt(
            np.asarray(factorial(absn-absm, exact=True) / factorial(absn+absm, exact=True), dtype=float))
        self.a = np.sqrt((absn+1+absm) * (absn+1-absm) / ((2*absn+1)*(2*absn+3)))
        self.b = np.sqrt((n-m-1) * (n-m) / ((2*n-1)*(2*n+1)))
        self.b[m<0] *= -1
        self.d = 0.5 * np.sqrt((n-m) * (n+m+1))
        self.d[m<0] *= -1
        with np.errstate(divide='ignore', invalid='ignore'):
            self.g = 2*(m+1) / np.sqrt((n-m)*(n+m+1))
            self.h = np.sqrt((n+m+2)*(n-m-1) / ((n-m)*(n+m+1)))
        self.absm = absm
        if not (
            np.all(np.isfinite(self.sqrt_factorial_ratio)) and
            np.all(np.isfinite(self.a)) and
            np.all(np.isfinite(self.b)) and
            np.all(np.isfinite(self.d)) and
            np.all(np.isfinite(self.absm))
        ):
            raise ValueError("Found a non-finite value inside this object")

    def workspace(self, cosβ):
        """Return a new workspace sized for cosβ."""
        return np.zeros((self.nmpm_total_size_plus1,) + cosβ.shape, dtype=float)

    def __call__(self, cosβ, workspace=None):
        cosβ = np.asarray(cosβ, dtype=float)
        if np.max(cosβ) > 1.0 or np.min(cosβ) < -1.0:
            raise ValueError('Nonsensical value for range of cosβ: [{0}, {1}]'.format(np.min(cosβ), np.max(cosβ)))
        cosβshape = cosβ.shape
        Hnmpm = workspace if workspace is not None else self.workspace(cosβ)
        cosβ = cosβ.ravel(order='K')
        sinβ = np.sqrt(1 - cosβ**2)
        Hnmpm = Hnmpm.reshape((-1, cosβ.size))
        _step_1(Hnmpm)
        _step_2(self.g, self.h, self.n_max, Hnmpm, cosβ, sinβ)
        _step_3(self.a, self.b, self.n_max, Hnmpm, cosβ, sinβ)
        _step_4(self.d, self.n_max, Hnmpm)
        _step_5(self.d, self.n_max, Hnmpm)
        _step_6(self.n_max, Hnmpm)
        Hnmpm.reshape((-1,)+cosβshape)
        return Hnmpm[:self.nmpm_total_size]  # Remove n_max+1 scratch space



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Modes class.  In
### particular, they assume that the first argument, `self` is an instance of Modes.  They should
### probably not be used outside of that class.

import copy
import numpy as np
from .. import LM_total_size
from ..multiplication import _multiplication_helper


def __array_ufunc__(self, ufunc, method, *args, out=None, **kwargs):
    # These are required for basic support, but can be more-or-less passed through because they return bools
    if ufunc in [np.not_equal, np.equal, np.logical_and, np.logical_or, np.isfinite, np.isinf, np.isnan]:
        args = [arg.view(np.ndarray) if isinstance(arg, type(self)) else arg for arg in args]
        kwargs['out'] = out
        return super(type(self), self).__array_ufunc__(ufunc, method, *args, **kwargs)

    # Here are all the functions we will support directly; all other ufuncs are probably meaningless
    elif ufunc not in [np.positive, np.negative, np.add, np.subtract,
                       np.multiply, np.divide, np.true_divide,
                       np.conj, np.conjugate, np.absolute]:
        return NotImplemented

    # We will not be supporting any more ufunc keywords
    if kwargs:
        raise NotImplementedError(f"Unrecognized arguments to {type(self).__name__}.__array_ufunc__: {kwargs}")

    if ufunc in [np.positive, np.negative]:
        out_view = out if out is None else out[0].view(np.ndarray)
        result = type(self)(ufunc(self.view(np.ndarray), out=out_view), **self._metadata)
        if out is not None and isinstance(out[0], type(self)):
            out[0]._metadata = copy.copy(self._metadata)

    elif ufunc in [np.add, np.subtract]:
        if isinstance(args[0], type(self)) and isinstance(args[1], type(self)):
            m1, m2 = args[:2]
            if m1.s != m2.s:
                return NotImplemented
            s = m1.s
            ell_min = min(m1.ell_min, m2.ell_min)
            ell_max = max(m1.ell_max, m2.ell_max)
            shape = np.broadcast(m1[..., 0], m2[..., 0]).shape + (LM_total_size(ell_min, ell_max),)
            result = out or np.zeros(shape, dtype=np.complex_)
            if isinstance(result, tuple):
                result = result[0]
            if isinstance(result, type(self)):
                result = result.view(np.ndarray)
            i_s1 = LM_total_size(ell_min, m1.ell_min-1)
            i_s2 = i_s1+LM_total_size(m1.ell_min, m1.ell_max)
            i_o1 = LM_total_size(ell_min, m2.ell_min-1)
            i_o2 = i_o1+LM_total_size(m2.ell_min, m2.ell_max)
            result[..., i_s1:i_s2] = m1.view(np.ndarray)
            if ufunc is np.subtract:
                result[..., i_o1:i_o2] -= m2.view(np.ndarray)
            else:
                result[..., i_o1:i_o2] += m2.view(np.ndarray)
            metadata = copy.copy(self._metadata)
            metadata['spin_weight'] = s
            metadata['ell_min'] = ell_min
            metadata['ell_max'] = ell_max
            result = type(self)(result, **metadata)
            if out is not None and isinstance(out[0], type(self)):
                metadata.pop('ell_min')
                out[0]._metadata = metadata
        elif isinstance(args[0], type(self)):
            modes = args[0]
            scalars = np.asanyarray(args[1])
            if np.any(scalars) or not modes._check_broadcasting(scalars):
                return NotImplemented
            result = type(self)(ufunc(modes.view(np.ndarray), scalars[..., np.newaxis], out=out), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = copy.copy(self._metadata)
        elif isinstance(args[1], type(self)):
            scalars = np.asanyarray(args[0])
            modes = args[1]
            if np.any(scalars) or not modes._check_broadcasting(scalars, reverse=True):
                return NotImplemented
            result = type(self)(ufunc(scalars[..., np.newaxis], modes.view(np.ndarray), out=out), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = copy.copy(self._metadata)
        else:
            return NotImplemented

    elif ufunc is np.multiply:
        if isinstance(args[0], type(self)) and isinstance(args[1], type(self)):
            s = args[0].view(np.ndarray)
            o = args[1].view(np.ndarray)
            result_s = args[0].s + args[1].s
            result_ell_min = 0
            result_ell_max = max(
                truncator((args[0].ell_max, args[1].ell_max))
                for truncator in [
                    args[0]._metadata.get('multiplication_truncator', sum),
                    args[1]._metadata.get('multiplication_truncator', sum)
                ]
            )
            result_shape = np.broadcast(s[..., 0], o[..., 0]).shape + (LM_total_size(result_ell_min, result_ell_max),)
            result = out or np.zeros(result_shape, dtype=np.complex_)
            if isinstance(result, tuple):
                result = result[0]
            if isinstance(result, type(self)):
                result = result.view(np.ndarray)
            _multiplication_helper(s, args[0].ell_min, args[0].ell_max, args[0].s,
                                   o, args[1].ell_min, args[1].ell_max, args[1].s,
                                   result, result_ell_min, result_ell_max, result_s)
            metadata = copy.copy(self._metadata)
            metadata['spin_weight'] = result_s
            metadata['ell_min'] = result_ell_min
            metadata['ell_max'] = result_ell_max
            result = type(self)(result, **metadata)
            if out is not None and isinstance(out[0], type(self)):
                metadata.pop('ell_min')
                out[0]._metadata = metadata
        elif isinstance(args[0], type(self)):
            modes = args[0]
            scalars = np.asanyarray(args[1])
            if not modes._check_broadcasting(scalars):
                return NotImplemented
            result = type(self)(ufunc(modes.view(np.ndarray), scalars[..., np.newaxis], out=out), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = copy.copy(self._metadata)
        elif isinstance(args[1], type(self)):
            scalars = np.asanyarray(args[0])
            modes = args[1]
            if not modes._check_broadcasting(scalars, reverse=True):
                return NotImplemented
            result = type(self)(ufunc(scalars[..., np.newaxis], modes.view(np.ndarray), out=out), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = copy.copy(self._metadata)
        else:
            return NotImplemented

    elif ufunc in [np.divide, np.true_divide]:
        if isinstance(args[1], type(self)) or not isinstance(args[0], type(self)):
            return NotImplemented
        modes = args[0]
        scalars = np.asanyarray(args[1])
        if not modes._check_broadcasting(scalars):
            return NotImplemented
        result = type(self)(ufunc(modes.view(np.ndarray), scalars[..., np.newaxis], out=out), **self._metadata)
        if out is not None and isinstance(out[0], type(self)):
            out[0]._metadata = copy.copy(self._metadata)

    elif ufunc in [np.conj, np.conjugate]:
        if isinstance(args[0], type(self)):
            s = args[0].view(np.ndarray)
            c = np.zeros_like(s) if out is None else out[0]
            for ell in range(abs(args[0].s), args[0].ell_max+1):
                i = LM_index(ell, 0, args[0].ell_min)
                if args[0].s%2 == 0:
                    c[..., i] = np.conjugate(s[..., i])
                else:
                    c[..., i] = -np.conjugate(s[..., i])
                for m in range(1, ell+1):
                    i_p, i_n = LM_index(ell, m, args[0].ell_min), LM_index(ell, -m, args[0].ell_min)
                    if (args[0].s+m)%2 == 0:
                        c[..., i_p], c[..., i_n] = np.conjugate(s[..., i_n]), np.conjugate(s[..., i_p])
                    else:
                        c[..., i_p], c[..., i_n] = -np.conjugate(s[..., i_n]), -np.conjugate(s[..., i_p])
            metadata = copy.copy(args[0]._metadata)
            metadata['spin_weight'] = -args[0].s
            result = type(self)(c, **metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = metadata
        else:
            return NotImplemented

    elif ufunc is np.absolute:
        return args[0].norm()

    else:
        # I thought we filtered everything else out...
        raise NotImplementedError(f"{type(self).__name__}.__array_ufunc__ has reached a point it should not have for ufunc {ufunc}")

    if result is NotImplemented:
        return NotImplemented

    if method == 'at':
        return

    return result



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Grid class.  In
### particular, they assume that the first argument, `self` is an instance of Grid.  They should
### probably not be used outside of that class.

import copy
import numpy as np


def conjugate(self, inplace=False):
    """Return Grid object corresponding to conjugated function

    The result has spin weight equal to the negative of this object's spin weight.

    """
    s = self.view(np.ndarray)
    c = s if inplace else np.empty_like(s)
    np.conjugate(s, out=c)
    if inplace:
        self._metadata['spin_weight'] = -self.s
        return self
    metadata = copy.copy(self._metadata)
    metadata['spin_weight'] = -self.s
    return type(self)(c, **metadata)


@property
def bar(self):
    return self.conjugate()
bar.__doc__ = conjugate.__doc__


@property
def real(self):
    """Return real-valued part of function values on the grid

    This only makes sense when the function has spin weight 0; anything else will raise an error.

    """
    if self.s != 0:
        raise ValueError("The real part of a function with non-zero spin weight is meaningless")
    s = self.view(np.ndarray)
    c = s.real
    return type(self)(c, **self._metadata)


@property
def imag(self):
    """Return imaginary-valued part of function values on the grid

    This only makes sense when the function has spin weight 0; anything else will raise an error.

    """
    if self.s != 0:
        raise ValueError("The imaginary part of a function with non-zero spin weight is meaningless")
    s = self.view(np.ndarray)
    c = s.imag
    return type(self)(c, **self._metadata)


def absolute(self):
    """Return absolute value of function values on the grid"""
    metadata = copy.copy(self._metadata)
    metadata['spin_weight'] = 0
    return type(self)(np.abs(self.view(np.ndarray)), **metadata)


def add(self, other):
    if isinstance(other, type(self)):
        s = self.s
        if s != other.s:
            raise ValueError(f"Cannot add functions with different spin weights ({s} and {other.s})")
        if self.n_theta != other.n_theta or self.n_phi != other.n_phi:
            raise ValueError(f"Shape mismatch: self.shape={self.shape}; other.shape={other.shape}")
        result = self.view(np.ndarray) + other.view(np.ndarray)
        return type(self)(result, **self._metadata)
    elif self.s != 0 and np.any(other):
        raise ValueError(f"It is not permitted to add non-zero scalars to a {type(self).__name__} object of non-zero spin weight")
    else:
        result = self.view(np.ndarray) + other
        return type(self)(result, **self._metadata)


def subtract(self, other):
    if isinstance(other, type(self)):
        if self.s != other.s:
            raise ValueError(f"Cannot subtract functions with different spin weights ({s} and {other.s})")
        if self.n_theta != other.n_theta or self.n_phi != other.n_phi:
            raise ValueError(f"Shape mismatch: self.shape={self.shape}; other.shape={other.shape}")
        result = self.view(np.ndarray) - other.view(np.ndarray)
        return type(self)(result, **self._metadata)
    elif self.s != 0 and np.any(other):
        raise ValueError(f"It is not permitted to subtract non-zero scalars from a {type(self).__name__} object of non-zero spin weight")
    else:
        result = self.view(np.ndarray) - other
        return type(self)(result, **self._metadata)


def multiply(self, other, truncator=None):
    """Multiply by another spin-weighted function or a scalar

    For spin-weighted functions, the spin weight of their product is the sum of the spin weights of
    the input.

    Parameters
    ==========
    other: Grid, array_like, complex, or float
        Grid object representing the spin-weighted functions, or an array or float which is
        equivalent to a spin-0 function.

    """
    if isinstance(other, type(self)):
        if self.n_theta != other.n_theta or self.n_phi != other.n_phi:
            raise ValueError(f"Shape mismatch: self.shape={self.shape}; other.shape={other.shape}")
        result = self.view(np.ndarray) * other.view(np.ndarray)
        result_s = self.s + other.s
        metadata = copy.copy(self._metadata)
        metadata['spin_weight'] = result_s
        return type(self)(result, **metadata)
    else:
        if self._check_broadcasting(other):
            return type(self)(self.view(np.ndarray) * other, **self._metadata)
        elif self._check_broadcasting(other, reverse=True):
            return type(self)(other * self.view(np.ndarray), **self._metadata)
        else:
            raise ValueError(f"Cannot broadcast input array to this {type(self).__name__} object.  Note that the input array\n           "
                             "must broadcast to all but last two dimensions of this object; it is not allowed to\n           "
                             "multiply each grid value individually.  If you really want to hack this, view this\n           "
                             "object as an ndarray, and don't complain if your results are wrong.")


def divide(self, other):
    """Divide by another spin-weighted function or a scalar

    For spin-weighted functions, the spin weight of their ratio is the difference of the spin
    weights of the input.

    Parameters
    ==========
    other: Grid, array_like, complex, or float
        Grid object representing the spin-weighted functions, or an array or float which is
        equivalent to a spin-0 function.

    """
    if isinstance(other, type(self)):
        if self.n_theta != other.n_theta or self.n_phi != other.n_phi:
            raise ValueError(f"Shape mismatch: self.shape={self.shape}; other.shape={other.shape}")
        result = self.view(np.ndarray) / other.view(np.ndarray)
        result_s = self.s - other.s
        metadata = copy.copy(self._metadata)
        metadata['spin_weight'] = result_s
        return type(self)(result, **metadata)
    else:
        if self._check_broadcasting(other):
            return type(self)(self.view(np.ndarray) / other, **self._metadata)
        else:
            raise ValueError(f"Cannot broadcast input array to this {type(self).__name__} object.  Note that the input array\n           "
                             "must broadcast to all but last two dimensions of this object; it is not allowed to\n           "
                             "multiply each grid value individually.  If you really want to hack this, view this\n           "
                             "object as an ndarray, and don't complain if your results are wrong.")



from spherical_functions import ell_max

# The coefficients were originally produced with the following code,
# though obviously this doesn't need to be run each time.

import numpy as np
import mpmath

mpmath.mp.dps = 4 * ell_max

_binomial_coefficients = np.empty((((2 * ell_max + 1) * (2 * ell_max + 2)) // 2,), dtype=float)
i = 0
for n in range(2 * ell_max + 1):
    for k in range(n + 1):
        _binomial_coefficients[i] = float(mpmath.binomial(n, k))
        i += 1
print(i, _binomial_coefficients.shape)
np.save('binomial_coefficients', _binomial_coefficients)

_ladder_operator_coefficients = np.empty((((2*ell_max + 3) * ell_max + 1),), dtype=float)
i = 0
for twoell in range(2*ell_max + 1):
    for twom in range(-twoell, twoell + 1, 2):
        _ladder_operator_coefficients[i] = mpmath.sqrt((twoell * (twoell + 2) - twom * (twom + 2))/4)
        i += 1
print(i, _ladder_operator_coefficients.shape)
np.save('ladder_operator_coefficients', _ladder_operator_coefficients)

# _Wigner_coefficients = np.empty(((4 * ell_max ** 3 + 12 * ell_max ** 2 + 11 * ell_max + 3) // 3,), dtype=float)
# i = 0
# for ell in range(ell_max + 1):
#     for mp in range(-ell, ell + 1):
#         for m in range(-ell, ell + 1):
#             rho_min = max(0, mp - m)
#             _Wigner_coefficients[i] = float(mpmath.sqrt(mpmath.fac(ell + m) * mpmath.fac(ell - m)
#                                                         / (mpmath.fac(ell + mp) * mpmath.fac(ell - mp)))
#                                             * mpmath.binomial(ell + mp, rho_min)
#                                             * mpmath.binomial(ell - mp, ell - m - rho_min))
#             i += 1

_Wigner_coefficients = np.empty(((8 * ell_max ** 3 + 18 * ell_max ** 2 + 13 * ell_max + 3) // 3,), dtype=float)
i = 0
for twoell in range(0, 2*ell_max + 1):
    for twomp in range(-twoell, twoell + 1, 2):
        for twom in range(-twoell, twoell + 1, 2):
            tworho_min = max(0, twomp - twom)
            _Wigner_coefficients[i] = float(mpmath.sqrt(mpmath.fac((twoell + twom)//2) * mpmath.fac((twoell - twom)//2)
                                                        / (mpmath.fac((twoell + twomp)//2) * mpmath.fac((twoell - twomp)//2)))
                                            * mpmath.binomial((twoell + twomp)//2, tworho_min//2)
                                            * mpmath.binomial((twoell - twomp)//2, (twoell - twom - tworho_min)//2))
            i += 1
print(i, _Wigner_coefficients.shape)
np.save('Wigner_coefficients', _Wigner_coefficients)



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Grid class.  In
### particular, they assume that the first argument, `self` is an instance of Grid.  They should
### probably not be used outside of that class.

import copy
import numpy as np


def __array_ufunc__(self, ufunc, method, *args, out=None, **kwargs):
    # These are required for basic support, but can be more-or-less passed through because they return bools
    if ufunc in [np.greater, np.greater_equal, np.less, np.less_equal, np.not_equal, np.equal,
                 np.logical_and, np.logical_or, np.isfinite, np.isinf, np.isnan]:
        args = [arg.view(np.ndarray) if isinstance(arg, type(self)) else arg for arg in args]
        kwargs['out'] = out
        return super(type(self), self).__array_ufunc__(ufunc, method, *args, **kwargs)

    # Here are all the functions we will support directly; all other ufuncs are probably meaningless
    elif ufunc not in [np.positive, np.negative, np.add, np.subtract,
                       np.multiply, np.divide, np.true_divide,
                       np.conj, np.conjugate, np.absolute,
                       np.power, np.sqrt, np.square, np.reciprocal]:
        return NotImplemented

    # We will not be supporting any more ufunc keywords
    if kwargs:
        raise NotImplementedError(f"Unrecognized arguments to {type(self).__name__}.__array_ufunc__: {kwargs}")

    if ufunc in [np.positive, np.negative]:
        out_view = out if out is None else out[0].view(np.ndarray)
        result = type(self)(ufunc(self.view(np.ndarray), out=out_view), **self._metadata)
        if out is not None and isinstance(out[0], type(self)):
            out[0]._metadata = copy.copy(self._metadata)

    elif ufunc in [np.add, np.subtract]:
        if isinstance(args[0], type(self)) and isinstance(args[1], type(self)):
            g1, g2 = args[:2]
            if g1.s != g2.s:
                raise ValueError(f"Cannot {ufunc.__name__} grids with different spin weights ({g1.s}, {g2.s})")
            if g1.n_theta != g2.n_theta or g1.n_phi != g2.n_phi:
                raise ValueError(f"Shape mismatch: grid1.shape={g1.shape}; grid2.shape={g2.shape}")
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(g1.view(np.ndarray), g2.view(np.ndarray), out=out_view), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                metadata = copy.copy(self._metadata)
                out[0]._metadata = metadata
        elif isinstance(args[0], type(self)):
            grid = args[0]
            scalars = np.asanyarray(args[1])
            if (grid.s!=0 and np.any(scalars)) or not grid._check_broadcasting(scalars):
                return NotImplemented
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(grid.view(np.ndarray), scalars[..., np.newaxis, np.newaxis], out=out_view), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = copy.copy(self._metadata)
        elif isinstance(args[1], type(self)):
            grid = args[1]
            scalars = np.asanyarray(args[0])
            if (grid.s!=0 and np.any(scalars)) or not grid._check_broadcasting(scalars, reverse=True):
                return NotImplemented
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(scalars[..., np.newaxis, np.newaxis], grid.view(np.ndarray), out=out_view), **self._metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = copy.copy(self._metadata)
        else:
            return NotImplemented

    elif ufunc in [np.multiply, np.divide, np.true_divide]:
        if isinstance(args[0], type(self)) and isinstance(args[1], type(self)):
            g1, g2 = args[:2]
            if g1.n_theta != g2.n_theta or g1.n_phi != g2.n_phi:
                raise ValueError(f"Shape mismatch: grid1.shape={g1.shape}; grid2.shape={g2.shape}")
            result_s = g1.s + g2.s if ufunc is np.multiply else g1.s - g2.s
            result_metadata = copy.copy(g1._metadata)
            result_metadata['spin_weight'] = result_s
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(g1.view(np.ndarray), g2.view(np.ndarray), out=out_view), **result_metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = result_metadata
        elif isinstance(args[0], type(self)):
            grid = args[0]
            scalars = np.asanyarray(args[1])
            if not grid._check_broadcasting(scalars):
                return NotImplemented
            result_s = grid.s
            result_metadata = copy.copy(grid._metadata)
            result_metadata['spin_weight'] = result_s
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(grid.view(np.ndarray), scalars[..., np.newaxis, np.newaxis], out=out_view), **result_metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = result_metadata
        elif isinstance(args[1], type(self)):
            grid = args[1]
            scalars = np.asanyarray(args[0])
            if not grid._check_broadcasting(scalars, reverse=True):
                return NotImplemented
            result_s = grid.s if ufunc is np.multiply else - grid.s
            result_metadata = copy.copy(grid._metadata)
            result_metadata['spin_weight'] = result_s
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(scalars[..., np.newaxis, np.newaxis], grid.view(np.ndarray), out=out_view), **result_metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = result_metadata
        else:
            return NotImplemented

    elif ufunc in [np.conj, np.conjugate]:
        if isinstance(args[0], type(self)):
            metadata = copy.copy(args[0]._metadata)
            metadata['spin_weight'] = -args[0].s
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(np.conjugate(args[0].view(np.ndarray), out=out_view), **metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = metadata
        else:
            return NotImplemented

    elif ufunc is np.absolute:
        if isinstance(args[0], type(self)):
            metadata = copy.copy(args[0]._metadata)
            metadata['spin_weight'] = 0
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(np.absolute(args[0].view(np.ndarray), out=out_view), **metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = metadata
        else:
            return NotImplemented

    elif ufunc is np.power:
        if isinstance(args[0], type(self)):
            try:
                exponent = int(args[1])
            except:
                return NotImplemented
            metadata = copy.copy(args[0]._metadata)
            metadata['spin_weight'] = exponent * args[0].s
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(np.power(args[0].view(np.ndarray), exponent, out=out_view), **metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = metadata
        else:
            return NotImplemented

    elif ufunc in [np.sqrt, np.square, np.reciprocal]:
        if isinstance(args[0], type(self)):
            if ufunc is np.sqrt:
                if args[0].s % 2 != 0:
                    return NotImplemented
                new_spin = args[0].s // 2
            elif ufunc is np.square:
                new_spin = args[0].s * 2
            elif ufunc is np.reciprocal:
                new_spin = -args[0].s
            metadata = copy.copy(args[0]._metadata)
            metadata['spin_weight'] = new_spin
            out_view = out if out is None else out[0].view(np.ndarray)
            result = type(self)(ufunc(args[0].view(np.ndarray), out=out_view), **metadata)
            if out is not None and isinstance(out[0], type(self)):
                out[0]._metadata = metadata
        else:
            return NotImplemented

    else:
        # I thought we filtered everything else out...
        raise NotImplementedError(f"{type(self).__name__}.__array_ufunc__ has reached a point it should not have for ufunc {ufunc}")

    if result is NotImplemented:
        return NotImplemented

    if method == 'at':
        return

    return result



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Modes class.  In
### particular, they assume that the first argument, `self` is an instance of Modes.  They should
### probably not be used outside of that class.

import copy
import numpy as np


def Lsquared(self):
    """Total angular-momentum operator

    This is the standard L^2 operator, familiar from basic physics, extended to work with SWSHs.
    It is also known as the Casimir operator, and is equal to

        L^2 = LxLx + LyLy + LzLz = 0.5(L+L- + L-L+) + LzLz

    Note that these are the left Lie derivatives, but L^2 = R^2, where R is the right Lie
    derivative.

    The left Lie derivative of a function f(Q) over the unit quaternions with respect to a
    generator of rotation g is defined as

        Lg(f){Q} = -0.5j df{exp(t*g) * Q} / dt |t=0

    This agrees with the usual angular-momentum operators familiar from spherical-harmonic
    theory, and reduces to it when the function has spin weight 0, but also applies to functions
    of general spin weight.  In terms of the SWSHs, we can write the action of Lsquared as

        Lsquared {s}Y{l,m} = l * (l+1) * {s}Y{l,m}

    """
    import numpy as np
    d = self.copy()
    s = self.view(np.ndarray)
    for ell in range(abs(self.s), self.ell_max+1):
        d[..., d.index(ell, -ell):d.index(ell, ell)+1] = (
            ell * (ell+1)
            * s[..., d.index(ell, -ell):d.index(ell, ell)+1]
        )
    return d


def Lz(self):
    """Left Lie derivative with respect to rotation about z

    The left Lie derivative of a function f(Q) over the unit quaternions with respect to a
    generator of rotation g is defined as

        Lg(f){Q} = -0.5j df{exp(t*g) * Q} / dt |t=0

    This agrees with the usual angular-momentum operators familiar from spherical-harmonic
    theory, and reduces to it when the function has spin weight 0, but also applies to functions
    of general spin weight.  In terms of the SWSHs, we can write the action of Lz as

        Lz {s}Y{l,m} = m * {s}Y{l,m}

    """
    import numpy as np
    d = self.copy()
    s = self.view(np.ndarray)
    for ell in range(abs(self.s), self.ell_max+1):
        for m in range(-ell, ell+1):
            d[..., d.index(ell, m)] = m * s[..., self.index(ell, m)]
    return d


def Lplus(self):
    """Raising operator for Lz

    We define Lplus to be the raising operator for the left Lie derivative with respect to
    rotation about z, Lz.  By definition, this means that [Lz, Lplus] = Lplus, which allows us
    to derive Lplus = Lx + 1j * Ly.  In terms of the SWSHs, we can write the action of Lplus as

        Lplus {s}Y{l,m} = sqrt((l-m)*(l+m+1)) {s}Y{l,m+1}

    Consequently, the modes of a function are affected as

        {Lplus f}{s, l, m} = sqrt((l+m)*(l-m+1)) * f{s,l,m-1}

    """
    # sYlm = (-1)**s sqrt((2ell+1)/(4pi)) D{l,m,-s}
    # Lplus {s}Y{l,m} = (-1)**s sqrt((2ell+1)/(4pi)) Lplus D{l,m,-s}
    #                 = (-1)**s sqrt((2ell+1)/(4pi)) sqrt((l-m)(l+m+1)) D{l,m+1,-s}
    #                 = sqrt((l-m)(l+m+1)) (-1)**s sqrt((2ell+1)/(4pi)) D{l,m+1,-s}
    #                 = sqrt((l-m)(l+m+1)) {s}Y{l,m+1}
    # {L+ f}{s', l', m'}
    #    = integral(L+ f {s'}Ybar{l',m'})  # Integral over rotation group
    #    = integral(L+ sum(f{s,l,m}{s}Y{l,m}) {s'}Ybar{l',m'})
    #    = sum(f{s,l,m} integral(L+ {s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(f{s,l,m} integral(sqrt((l-m)(l+m+1)) {s}Y{l,m+1} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l-m)(l+m+1)) f{s,l,m} integral({s}Y{l,m+1} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l-m)(l+m+1)) f{s,l,m} delta{s, s'} delta{m+1, m'} delta{l, l'}
    #    = sqrt((l'-(m'-1))(l'+(m'-1)+1)) f{s',l',m'-1}
    #    = sqrt((l'+m')(l'-m'+1)) f{s',l',m'-1}
    # {L+ f}{s, l, m} = sqrt((l+m)(l-m+1)) f{s,l,m-1}
    import math
    import numpy as np
    d = np.zeros_like(self)
    s = self.view(np.ndarray)
    for ell in range(abs(self.s), self.ell_max+1):
        d[..., self.index(ell, -ell)] = 0.0
        for m in range(-ell+1, ell+1):
            d[..., d.index(ell, m)] = math.sqrt((ell+m)*(ell-m+1)) * s[..., self.index(ell, m-1)]
    return d


def Lminus(self):
    """Lowering operator for Lz

    We define Lminus to be the lowering operator for the left Lie derivative with respect to
    rotation about z, Lz.  By definition, this means that [Lz, Lminus] = -Lminus, which allows
    us to derive Lminus = Lx - 1j * Ly.  In terms of the SWSHs, we can write the action of
    Lminus as

        Lminus {s}Y{l,m} = sqrt((l+m)*(l-m+1)) * {s}Y{l,m-1}

    Consequently, the modes of a function are affected as

        {Lminus f}{s, l, m} = sqrt((l-m)*(l+m+1)) * f{s,l,m+1}

    """
    # sYlm = (-1)**s sqrt((2ell+1)/(4pi)) D{l,m,-s}
    # Lminus {s}Y{l,m} = (-1)**s sqrt((2ell+1)/(4pi)) Lminus D{l,m,-s}
    #                  = (-1)**s sqrt((2ell+1)/(4pi)) sqrt((l+m)(l-m+1)) D{l,m-1,-s}
    #                  = sqrt((l+m)(l-m+1)) (-1)**s sqrt((2ell+1)/(4pi)) D{l,m-1,-s}
    #                  = sqrt((l+m)(l-m+1)) {s}Y{l,m-1}
    # {L- f}{s', l', m'}
    #    = integral(L- f {s'}Ybar{l',m'})  # Integral over rotation group
    #    = integral(L- sum(f{s,l,m}{s}Y{l,m}) {s'}Ybar{l',m'})
    #    = sum(f{s,l,m} integral(L- {s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(f{s,l,m} integral(sqrt((l+m)(l-m+1)) {s}Y{l,m-1} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l+m)(l-m+1)) f{s,l,m} integral({s}Y{l,m-1} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l+m)(l-m+1)) f{s,l,m} delta{s, s'} delta{m-1, m'} delta{l, l'}
    #    = sqrt((l'+(m'+1))(l'-(m'+1)+1)) f{s',l',m'+1}
    #    = sqrt((l'-m')(l'+m'+1)) f{s',l',m'+1}
    # {L- f}{s, l, m} = sqrt((l-m)(l+m+1)) f{s,l,m+1}
    import math
    import numpy as np
    d = np.zeros_like(self)
    s = self.view(np.ndarray)
    for ell in range(abs(self.s), self.ell_max+1):
        for m in range(-ell, ell):
            d[..., self.index(ell, m)] = math.sqrt((ell-m)*(ell+m+1)) * s[..., self.index(ell, m+1)]
        d[..., self.index(ell, ell)] = 0.0
    return d


def Rsquared(self):
    """Total angular-momentum operator

    This is the R^2 operator, much like the L^2 operator familiar from basic physics, but using
    the right Lie derivative, and extended to work with SWSHs.  It is also known as the Casimir
    operator, and is equal to

        R^2 = RxRx + RyRy + RzRz = 0.5(R+R- + R-R+) + RzRz

    Note that these are the right Lie derivatives, but L^2 = R^2, where L is the left Lie
    derivative.

    The right Lie derivative of a function f(Q) over the unit quaternions with respect to a
    generator of rotation g is defined as

        Rg(f){Q} = -0.5j df{Q * exp(t*g)} / dt |t=0

    This is unlike the usual angular-momentum operators Lz, etc., familiar from
    spherical-harmonic theory because the exponential is on the right-hand side of the argument.
    This operator is less common in physics because it represents the dependence of the function
    on the choice of frame.  In terms of the SWSHs, we can write the action of Rsquared as

        Rsquared {s}Y{l,m} = l * (l+1) * {s}Y{l,m}

    """
    return self.Lsquared()


def Rz(self):
    """Right Lie derivative with respect to rotation about z

    The right Lie derivative of a function f(Q) over the unit quaternions with respect to a
    generator of rotation g is defined as

        Rg(f){Q} = -0.5j df{Q * exp(t*g)} / dt |t=0

    This is unlike the usual angular-momentum operators Lz, etc., familiar from
    spherical-harmonic theory because the exponential is on the right-hand side of the argument.
    This operator is less common in physics because it represents the dependence of the function
    on the choice of frame.  In terms of the SWSHs, we can write the action of Rz as

        Rz {s}Y{l,m} = -s * {s}Y{l,m}

    Equivalently, the modes of a function are affected as

        {Rz f} {s,l,m} = -s * f{s,l,m}

    Note the unfortunate sign of `s`, which seems to be opposite to what we expect, and arises
    from the choice of definition of `s` in the original paper by Newman and Penrose.

    """
    # {Rzf}{s', l', m'}
    #    = integral(Rz f {s'}Ybar{l',m'})  # Integral over rotation group
    #    = integral(Rz sum(f{s,l,m}{s}Y{l,m}) {s'}Ybar{l',m'})
    #    = sum(f{s,l,m} integral(Rz {s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(f{s,l,m} integral(-s {s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(-s f{s,l,m} integral({s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(-s f{s,l,m} delta{s, s'} delta{m, m'} delta{l, l'}
    #    = -s f{s',l',m'}
    # {Rzf}{s, l, m} = -s f{s,l,m}
    import numpy as np
    return type(self)(-self.s * self.view(np.ndarray), **self._metadata)


def Rplus(self):
    """Raising operator for Rz

    We define Rplus to be the raising operator for the right Lie derivative with respect to
    rotation about z, Rz.  By definition, this means that [Rz, Rplus] = Rplus, which allows us
    to derive Rplus = Rx - 1j * Ry.  In terms of the SWSHs, we can write the action of Rplus as

        Rplus {s}Y{l,m} = sqrt((l+s)(l-s+1)) {s-1}Y{l,m}

    Consequently, the modes of a function are affected as

        {Rplus f} {s,l,m} = sqrt((l-s)(l+s+1)) f{s+1,l,m}

    Again, because of the unfortunate choice of the sign of `s` made in the original paper by
    Newman and Penrose, this looks like a lowering operator for `s`.  But it really is a raising
    operator for Rz, and raises the eigenvalue of the corresponding Wigner matrix - though that
    lowers the value of `s`.

    """
    # sYlm = (-1)**s sqrt((2ell+1)/(4pi)) D{l,m,-s}
    # Rplus {s}Y{l,m} = (-1)**s sqrt((2ell+1)/(4pi)) Rplus D{l,m,-s}
    #                 = (-1)**s sqrt((2ell+1)/(4pi)) sqrt((l+s)(l-s+1)) D{l,m,-s+1}
    #                 = sqrt((l+s)(l-s+1)) (-1)**s sqrt((2ell+1)/(4pi)) D{l,m,-s+1}
    #                 = sqrt((l+s)(l-s+1)) {s-1}Y{l,m}
    # {R+f}{s', l', m'}
    #    = integral(R+ f {s'}Ybar{l',m'})  # Integral over rotation group
    #    = integral(R+ sum(f{s,l,m}{s}Y{l,m}) {s'}Ybar{l',m'})
    #    = sum(f{s,l,m} integral(R+ {s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(f{s,l,m} integral(sqrt((l+s)(l-s+1)) {s-1}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l+s)(l-s+1)) f{s,l,m} integral({s-1}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l+s)(l-s+1)) f{s,l,m} delta{s-1, s'} delta{m, m'} delta{l, l'}
    #    = sqrt((l'+s'+1)(l'-(s'+1)+1) f{s'+1,l',m'}
    # {R+f}{s, l, m} = sqrt((l-s)(l+s+1)) f{s+1,l,m}
    import math
    import numpy as np
    metadata = copy.copy(self._metadata)
    metadata['spin_weight'] = self.s-1
    metadata['ell_min'] = min(abs(self.s-1), self.ell_min)
    metadata['ell_max'] = self.ell_max
    d = type(self)(np.zeros_like(self.view(np.ndarray)), **metadata)
    s = self.view(np.ndarray)
    for ell in range(max(abs(d.s), abs(self.s)), d.ell_max+1):
        if ell >= self.ell_min:
            d[..., d.index(ell, -ell):d.index(ell, ell)+1] = (
                math.sqrt((ell-d.s)*(ell+d.s+1))
                * s[..., self.index(ell, -ell):self.index(ell, ell)+1]
            )
    return d


def Rminus(self):
    """Lowering operator for Rz

    We define Rminus to be the lowering operator for the right Lie derivative with respect to
    rotation about z, Rz.  By definition, this means that [Rz, Rminus] = -Rminus, which allows
    us to derive Rminus = Rx + 1j * Ry.  In terms of the SWSHs, we can write the action of
    Rminus as

        Rminus {s}Y{l,m} = sqrt((l-s)(l+s+1)) {s+1}Y{l,m}

    Consequently, the modes of a function are affected as

        {Rminus f} {s,l,m} = sqrt((l+s)(l-s+1)) f{s-1,l,m}

    Again, because of the unfortunate choice of the sign of `s` made in the original paper by
    Newman and Penrose, this looks like a raising operator for `s`.  But it really is a lowering
    operator for Rz, and lowers the eigenvalue of the corresponding Wigner matrix - though that
    raises the value of `s`.

    """
    # sYlm = (-1)**s sqrt((2ell+1)/(4pi)) D{l,m,-s}
    # Rminus {s}Y{l,m} = (-1)**s sqrt((2ell+1)/(4pi)) Rminus D{l,m,-s}
    #                  = (-1)**s sqrt((2ell+1)/(4pi)) sqrt((l-s)(l+s+1)) D{l,m,-s-1}
    #                  = sqrt((l-s)(l+s+1)) (-1)**s sqrt((2ell+1)/(4pi)) D{l,m,-s-1}
    #                  = sqrt((l-s)(l+s+1)) {s+1}Y{l,m}
    # {R- f}{s', l', m'}
    #    = integral(R- f {s'}Ybar{l',m'})  # Integral over rotation group
    #    = integral(R- sum(f{s,l,m}{s}Y{l,m}) {s'}Ybar{l',m'})
    #    = sum(f{s,l,m} integral(R- {s}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(f{s,l,m} integral(sqrt((l-s)(l+s+1)) {s+1}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l-s)(l+s+1)) f{s,l,m} integral({s+1}Y{l,m} {s'}Ybar{l',m'}))
    #    = sum(sqrt((l-s)(l+s+1)) f{s,l,m} delta{s+1, s'} delta{m, m'} delta{l, l'}
    #    = sqrt((l'-(s'-1))(l'+(s'-1)+1)) f{s'-1,l',m'}
    #    = sqrt((l'-s'+1)(l'+s')) f{s'-1,l',m'}
    # {R- f}{s, l, m} = sqrt((l+s)(l-s+1)) f{s-1,l,m}
    import math
    import numpy as np
    metadata = copy.copy(self._metadata)
    metadata['spin_weight'] = self.s+1
    metadata['ell_min'] = min(abs(self.s+1), self.ell_min)
    metadata['ell_max'] = self.ell_max
    d = type(self)(np.zeros_like(self.view(np.ndarray)), **metadata)
    s = self.view(np.ndarray)
    for ell in range(max(abs(d.s), abs(self.s)), d.ell_max+1):
        if ell >= self.ell_min:
            d[..., d.index(ell, -ell):d.index(ell, ell)+1] = (
                math.sqrt((ell+d.s)*(ell-d.s+1))
                * s[..., self.index(ell, -ell):self.index(ell, ell)+1]
            )
    return d


@property
def eth(self):
    """Spin-raising derivative operator defined by Newman and Penrose

    Note that this is identical to `Rminus`.

    By definition, the spin-raising operator satisfies [S, eth] = eth (where S is the spin
    operator, which just multiplies the function by its spin).  In terms of the SWSHs, we can
    write the action of eth as

        eth {s}Y{l,m} = sqrt((l-s)(l+s+1)) {s+1}Y{l,m}

    Consequently, the modes of a function are affected as

        {eth f} {s,l,m} = sqrt((l+s)(l-s+1)) f{s-1,l,m}

    """
    return self.Rminus()


@property
def ethbar(self):
    """Spin-lowering conjugate-derivative operator defined by Newman and Penrose

    Note that this is identical to `Rplus`, except with an opposite sign, to follow the
    conventions laid out by Newman and Penrose.

    By definition, the spin-lowering operator satisfies [S, ethbar] = -ethbar (where S is the
    spin operator, which just multiplies the function by its spin).  In terms of the SWSHs, we
    can write the action of ethbar as

        ethbar {s}Y{l,m} = -sqrt((l+s)(l-s+1)) {s-1}Y{l,m}

    Consequently, the modes of a function are affected as

        {ethbar f} {s,l,m} = -sqrt((l-s)(l+s+1)) f{s+1,l,m}

    """
    return -self.Rplus()



# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/waveforms/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

from math import sqrt, pi
import numpy as np
from quaternion.numba_wrapper import jit, xrange
from . import LM_total_size


@jit
def constant_as_ell_0_mode(constant):
    """Express constant as Y_{0,0} mode weight"""
    return constant * sqrt(4 * pi)


@jit
def constant_from_ell_0_mode(modes):
    """Express Y_{0,0} mode as constant

    This is just the inverse of the `constant_as_ell_0_mode` function.  Note that this does not assume that the
    output constant should be real.  If you want it to be, you must take the real part manually.

    """
    return modes / sqrt(4 * pi)


@jit
def vector_as_ell_1_modes(vector):
    """Express vector as Y_{1,m} mode weights

    A vector can be represented as a linear combination of weights of the ell=1 scalar spherical harmonics.
    Explicitly, if nhat is the usual unit vector in the (theta, phi) direction, then we can define a function
      v(theta, phi) = vector . nhat
    where the `.` represents the dot product, and v(theta, phi) is a pure ell=1 function.  This function simply
    returns the weights of that representation.

    Parameter
    ---------
    vector : float array of length 3
        The input should be an iterable containing the [v_x, v_y, v_z] components of the vector in that order

    Returns
    -------
    float array
       The returned object contains the (1,-1), (1,0), and (1,1) modes in that order

    """
    return np.stack((np.asarray((vector[..., 0] + 1j * vector[..., 1]) * sqrt(2 * pi / 3.)),
                     np.asarray(vector[..., 2] * sqrt(4 * pi / 3.)),
                     np.asarray((-vector[..., 0] + 1j * vector[..., 1]) * sqrt(2 * pi / 3.))), axis=-1)


@jit
def vector_from_ell_1_modes(modes):
    """Express Y_{1,m} modes as vector

    This is just the inverse of the `vector_as_ell_1_modes` function.  Note that this does not assume that the vector
    should be real-valued.  If you want it to be, you must call the `.real` method of the output array manually.

    """
    return np.stack((np.asarray((modes[..., 0] - modes[..., 2]) / (2 * sqrt(2 * pi / 3.))),
                     np.asarray((modes[..., 0] + modes[..., 2]) / (2j * sqrt(2 * pi / 3.))),
                     np.asarray(modes[..., 1] / sqrt(4 * pi / 3.))), axis=-1)


@jit
def eth_GHP(modes, spin_weight, ell_min=0):
    """Spin-raising eth operator as defined by Geroch-Held-Penrose

    N.B.: For our purposes, eth_GHP is the same as eth_NP/sqrt(2).

    This operator is defined in J. Math. Phys. 14, 874 (1973) <http://link.aip.org/link/?JMP/14/874/1>. The eth
    operator was originally defined by Newman and Penrose in J. Math. Phys. 7, 863 (1966)
    <http://link.aip.org/link/?JMP/7/863/1>, and discussed further by Goldberg et al., J. Math. Phys. 8, 2155 (1967)
    <http://link.aip.org/link/?JMP/8/2155/1>.  In the case of a spacelike 2-surface (which is all we treat in this
    function), the GHP eth operator reduces almost to the original.  By comparing GHP's Eq. (3.8) to Eq. (2.13) of
    Goldberg et al., we can see the difference.  As GHP say below their Eq. (3.11) that "for complete agreement in
    the case of a general 2-surface metric we would have (strictly speaking) to multiply up our eth operator by a
    factor sqrt(2)."

    Parameters
    ----------
    modes : 1-d complex array
        This array contains the modes starting from ell=ell_min, and continuing in standard order (as in sf.LM_range).
    spin_weight : int
        Spin weight of the input field.  The eth operators raise the spin weight by 1.
    ell_min : int, optional
        Smallest ell value present in input `modes`.  Defaults to 0.

    Returns
    -------
    1-d complex array
        The output has the same size as the input `modes`, and corresponds to the same (ell,m) values.  Note, however,
        that these modes have spin weight greater than the input by 1.

    """
    ell_max = int(sqrt(len(modes) + LM_total_size(0, ell_min - 1))) - 1
    eth_modes = np.copy(modes)
    i_mode = 0
    for ell in xrange(ell_min, ell_max + 1):
        factor = (0.0 if ell < abs(spin_weight + 1) else sqrt((ell - spin_weight) * (ell + spin_weight + 1.) / 2.))
        for m in xrange(-ell, ell + 1):
            eth_modes[i_mode] *= factor
            i_mode += 1
    return eth_modes


@jit
def ethbar_GHP(modes, spin_weight, ell_min=0):
    """Spin-lowering \bar{eth} operator as defined by Geroch-Held-Penrose

    N.B.: For our purposes, eth_GHP is the same as eth_NP/sqrt(2).

    This is the complex conjugate of the `eth_GHP` operator.  See that function's docstring for more information.

    Parameters
    ----------
    modes : 1-d complex array
        This array contains the modes starting from ell=ell_min, and continuing in standard order (as in sf.LM_range).
    spin_weight : int
        Spin weight of the input field.  The \bar{eth} operators lower the spin weight by 1.
    ell_min : int, optional
        Smallest ell value present in input `modes`.  Defaults to 0.

    Returns
    -------
    1-d complex array
        The output has the same size as the input `modes`, and corresponds to the same (ell,m) values.  Note, however,
        that these modes have spin weight less than the input by 1.

    """
    ell_max = int(sqrt(len(modes) + LM_total_size(0, ell_min - 1))) - 1
    ethbar_modes = np.copy(modes)
    i_mode = 0
    for ell in xrange(ell_min, ell_max + 1):
        factor = (0.0 if ell < abs(spin_weight - 1) else -sqrt((ell + spin_weight) * (ell - spin_weight + 1.) / 2.))
        for m in xrange(-ell, ell + 1):
            ethbar_modes[i_mode] *= factor
            i_mode += 1
    return ethbar_modes


@jit
def eth_NP(modes, spin_weight, ell_min=0):
    """Spin-raising eth operator as defined by Newman and Penrose

    N.B.: For our purposes, eth_GHP is the same as eth_NP/sqrt(2).

    This is the original eth operator defined by Newman and Penrose.  See the documentation for eth_GHP for more detail.

    Parameters
    ----------
    modes : 1-d complex array
        This array contains the modes starting from ell=ell_min, and continuing in standard order (as in sf.LM_range).
    spin_weight : int
        Spin weight of the input field.  The eth operators raise the spin weight by 1.
    ell_min : int, optional
        Smallest ell value present in input `modes`.  Defaults to 0.

    Returns
    -------
    1-d complex array
        The output has the same size as the input `modes`, and corresponds to the same (ell,m) values.  Note, however,
        that these modes have spin weight greater than the input by 1.

    """
    ell_max = int(sqrt(len(modes) + LM_total_size(0, ell_min - 1))) - 1
    eth_modes = np.copy(modes)
    i_mode = 0
    for ell in xrange(ell_min, ell_max + 1):
        factor = (0.0 if ell < abs(spin_weight + 1) else sqrt((ell - spin_weight) * (ell + spin_weight + 1.)))
        for m in xrange(-ell, ell + 1):
            eth_modes[i_mode] *= factor
            i_mode += 1
    return eth_modes


@jit
def ethbar_NP(modes, spin_weight, ell_min=0):
    """Spin-lowering \bar{eth} operator as defined by Newman and Penrose

    N.B.: For our purposes, eth_GHP is the same as eth_NP/sqrt(2).

    This is the complex conjugate of the `eth_NP` operator.  See that function's docstring for more information.

    Parameters
    ----------
    modes : 1-d complex array
        This array contains the modes starting from ell=ell_min, and continuing in standard order (as in sf.LM_range).
    spin_weight : int
        Spin weight of the input field.  The \bar{eth} operators lower the spin weight by 1.
    ell_min : int, optional
        Smallest ell value present in input `modes`.  Defaults to 0.

    Returns
    -------
    1-d complex array
        The output has the same size as the input `modes`, and corresponds to the same (ell,m) values.  Note, however,
        that these modes have spin weight less than the input by 1.

    """
    ell_max = int(sqrt(len(modes) + LM_total_size(0, ell_min - 1))) - 1
    ethbar_modes = np.copy(modes)
    i_mode = 0
    for ell in xrange(ell_min, ell_max + 1):
        factor = (0.0 if ell < abs(spin_weight - 1) else -sqrt((ell + spin_weight) * (ell - spin_weight + 1.)))
        for m in xrange(-ell, ell + 1):
            ethbar_modes[i_mode] *= factor
            i_mode += 1
    return ethbar_modes


@jit
def ethbar_inverse_NP(modes, spin_weight, ell_min=0):
    """Inverse of the spin-lowering \bar{eth} operator as defined by Newman and Penrose

    This function acts as a (partial) inverse or integral of the `ethbar_NP` operator.  (See that function's
    docstring for more information, including the calling signature.)  In particular, composing the two functions in
    either order should give (almost) the identity function.  Essentially, this is an integral of the ethbar
    function; so this is only a partial inverse because constants of integration could be added in some cases.  The
    difference between this function and `eth_NP` is mostly in the normalization.

    """
    ell_max = int(sqrt(len(modes) + LM_total_size(0, ell_min - 1))) - 1
    ethbar_inverse_modes = np.copy(modes)
    i_mode = 0
    for ell in xrange(ell_min, ell_max + 1):
        term = (ell + spin_weight + 1.) * (ell - spin_weight)
        if term > 0.0:
            factor = -sqrt(term)
            for m in xrange(-ell, ell + 1):
                ethbar_inverse_modes[i_mode] /= factor
                i_mode += 1
        else:
            i_mode += 2*ell+1
    return ethbar_inverse_modes



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Modes class.  In
### particular, they assume that the first argument, `self` is an instance of Modes.  They should
### probably not be used outside of that class.

import copy
import numpy as np
from .. import LM_total_size, LM_index
from ..multiplication import _multiplication_helper


def conjugate(self, inplace=False):
    """Return Modes object corresponding to conjugated function

    The operations of conjugation and decomposition into mode weights do not commute.  That is,
    the modes of a conjugated function do not equal the conjugated modes of a function.  So,
    rather than simply returning the conjugate of the data from this Modes object, this function
    returns a Modes object containing the data for the conjugated function.

    If `inplace` is True, then the operation is performed in place, modifying this Modes object
    itself.  Note that some copying is still needed, but only for 2 modes at a time, and those
    copies are freed after this function returns.

    Here is the derivation:

        f = sum(f{s, l,m} * {s}Y{l,m} for l, m in LM)
        conjugate(f) = sum(conjugate(f{s, l,m}) * conjugate({s}Y{l,m}))
                     = sum(conjugate(f{s, l,m}) * (-1)**(s+m) * {-s}Y{l,-m})
                     = sum((-1)**(s+m) * conjugate(f{s, l, -m}) * {-s}Y{l,m})

        conjugate(f){s', l',m'} = integral(
            sum((-1)**(s+m) * conjugate(f{s,l,-m}) * {-s}Y{l,m}) * {s'}Y{l',m'},
            dR  # integration over rotation group
        )
        = sum((-1)**(s+m) * conjugate(f{s,l,-m}) * delta_{-s, s'} * delta_{l, l'} * delta_{m, m'})
        = (-1)**(s'+m') * conjugate(f{-s', l', -m'})

    The result is this:

        conjugate(f){s, l, m} = (-1)**(s+m) * conjugate(f{-s, l, -m})

    """
    s = self.view(np.ndarray)
    c = s if inplace else np.zeros_like(s)
    for ell in range(abs(self.s), self.ell_max+1):
        i = LM_index(ell, 0, self.ell_min)
        if self.s%2 == 0:
            c[..., i] = np.conjugate(s[..., i])
        else:
            c[..., i] = -np.conjugate(s[..., i])
        for m in range(1, ell+1):
            i_p, i_n = LM_index(ell, m, self.ell_min), LM_index(ell, -m, self.ell_min)
            if (self.s+m)%2 == 0:
                c[..., i_p], c[..., i_n] = np.conjugate(s[..., i_n]), np.conjugate(s[..., i_p])
            else:
                c[..., i_p], c[..., i_n] = -np.conjugate(s[..., i_n]), -np.conjugate(s[..., i_p])
    if inplace:
        self._metadata['spin_weight'] = -self.s
        return self
    metadata = copy.copy(self._metadata)
    metadata['spin_weight'] = -self.s
    return type(self)(c, **metadata)


@property
def bar(self):
    return self.conjugate()
bar.__doc__ = conjugate.__doc__


def _real_func(self, inplace=False):
    """Return Modes object corresponding to real-valued function

    Note that this only makes sense for functions of spin weight zero; other spins will raise
    ValueErrors.

    The condition that a function `f` be real is given by

        f{l, m} = conjugate(f){l, m} = (-1)**(m) * conjugate(f{l, -m})

    [Note that conj(f){l, m} != conj(f{l, m}).  See Modes.conjugate docstring.]

    We enforce that condition by essentially averaging the two modes:

        f{l, m} = (f{l, m} + (-1)**m * conjugate(f{l, -m})) / 2

    """
    if self.s != 0:
        raise ValueError("The real part of a function with non-zero spin weight is meaningless")
    s = self.view(np.ndarray)
    c = s if inplace else np.zeros_like(s)
    for ell in range(abs(self.s), self.ell_max+1):
        i = LM_index(ell, 0, self.ell_min)
        c[..., i] = np.real(s[..., i])
        for m in range(1, ell+1):
            i_p, i_n = LM_index(ell, m, self.ell_min), LM_index(ell, -m, self.ell_min)
            if m%2 == 0:
                c[..., i_p] = (s[..., i_p] + np.conjugate(s[..., i_n])) / 2
                c[..., i_n] = np.conjugate(c[..., i_p])
            else:
                c[..., i_p] = (s[..., i_p] - np.conjugate(s[..., i_n])) / 2
                c[..., i_n] = -np.conjugate(c[..., i_p])
    if inplace:
        return self
    return type(self)(c, **self._metadata)


@property
def real(self):
    return self._real_func(False)


def _imag_func(self, inplace=False):
    """Return Modes object corresponding to imaginary-valued function

    Note that this only makes sense for functions of spin weight zero; other spins will raise
    ValueErrors.

    The condition that a function `f` be imaginary is given by

        f{l, m} = -conjugate(f){l, m} = (-1)**(m+1) * conjugate(f{l, -m})

    [Note that conj(f){l, m} != conj(f{l, m}).  See Modes.conjugate docstring.]

    We enforce that condition by essentially averaging the two modes:

        f{l, m} = (f{l, m} - (-1)**m * conjugate(f{l, -m})) / 2

    """
    if self.s != 0:
        raise ValueError("The imaginary part of a function with non-zero spin weight is meaningless")
    s = self.view(np.ndarray)
    c = s if inplace else np.zeros_like(s)
    for ell in range(abs(self.s), self.ell_max+1):
        i = LM_index(ell, 0, self.ell_min)
        c[..., i] = 1j * np.imag(s[..., i])
        for m in range(1, ell+1):
            i_p, i_n = LM_index(ell, m, self.ell_min), LM_index(ell, -m, self.ell_min)
            if m%2 == 0:
                c[..., i_p] = (s[..., i_p] - np.conjugate(s[..., i_n])) / 2
                c[..., i_n] = -np.conjugate(c[..., i_p])
            else:
                c[..., i_p] = (s[..., i_p] + np.conjugate(s[..., i_n])) / 2
                c[..., i_n] = np.conjugate(c[..., i_p])
    if inplace:
        return self
    return type(self)(c, **self._metadata)


@property
def imag(self):
    return self._imag_func(False)


def norm(self):
    return np.linalg.norm(self.view(np.ndarray), axis=-1)


def add(self, other, subtraction=False):
    if isinstance(other, type(self)):
        if self.s != other.s:
            raise ValueError(f"Cannot add modes with different spin weights ({self.s} and {other.s})")
        s = self.s
        ell_min = min(self.ell_min, other.ell_min)
        ell_max = max(self.ell_max, other.ell_max)
        shape = np.broadcast(self[..., 0], other[..., 0]).shape + (LM_total_size(ell_min, ell_max),)
        result = np.zeros(shape, dtype=np.complex_)
        i_s1 = LM_total_size(ell_min, self.ell_min-1)
        i_s2 = i_s1+LM_total_size(self.ell_min, self.ell_max)
        i_o1 = LM_total_size(ell_min, other.ell_min-1)
        i_o2 = i_o1+LM_total_size(other.ell_min, other.ell_max)
        result[..., i_s1:i_s2] = self.view(np.ndarray)
        if subtraction:
            result[..., i_o1:i_o2] -= other.view(np.ndarray)
        else:
            result[..., i_o1:i_o2] += other.view(np.ndarray)
        metadata = copy.copy(self._metadata)
        metadata['spin_weight'] = s
        metadata['ell_min'] = ell_min
        metadata['ell_max'] = ell_max
        return type(self)(result, **metadata)
    elif np.any(other):
        raise ValueError(f"It is not permitted to add nonzero scalars to a {type(self).__name__} object")
    else:
        return np.add(self, other)


def subtract(self, other):
    if isinstance(other, type(self)):
        if self.s != other.s:
            raise ValueError(f"Cannot subtract modes with different spin weights ({self.s} and {other.s})")
        return self.add(other, True)
    elif np.any(other):
        raise ValueError(f"It is not permitted to add nonzero scalars to a {type(self).__name__} object")
    else:
        return np.subtract(self, other)


def multiply(self, other, truncator=None):
    """Multiply by another spin-weighted function or a scalar

    For spin-weighted functions, the spin weight of their product is the sum of the spin weights of
    the input.

    If those functions are band limited to maximum ell values, their product has maximum ell value
    given by the sum of the input maximum ell values.  Note that this output ell_max can be
    controlled by the `truncator` argument to this function; if the result is smaller than the sum
    of the input maximum ell values, this is equivalent to performing the full multiplication and
    then setting higher ell modes to zero.  The benefit of this type of truncation is that the
    higher modes don't even need to be computed, and no aliasing will result.

    Parameters
    ==========
    other: Modes, array_like, complex, or float
        Modes objects representing the spin-weighted functions, or an array or float which is
        equivalent to a spin-0 function.
    truncator: None or callable [defaults to None]
        Function to be applied to the tuple (self.ell_max, other.ell_max) to produce the ell_max for
        the resulting function.  Sensible values of the truncator include the built-in python
        functions `min`, `max`, and `sum` -- the latter giving the full "correct" answer.  If you
        want a specific ell value, you can use `lambda tup: max_ell`.  If None, this function falls
        back on the `multiplication_truncator` metadata fields of the input Modes objects, and uses
        the greater of the values that they return.  If either input object is missing the
        `multiplication_truncator` metadata field, it defaults to `sum`.

    """
    if isinstance(other, type(self)):
        s = self.view(np.ndarray)
        o = other.view(np.ndarray)
        new_s = self.s + other.s
        new_ell_min = 0
        if truncator is not None:
            new_ell_max = truncator((self.ell_max, other.ell_max))
        else:
            new_ell_max = max(
                truncator((self.ell_max, other.ell_max))
                for truncator in [
                    self._metadata.get('multiplication_truncator', sum),
                    other._metadata.get('multiplication_truncator', sum)
                ]
            )
        new_shape = np.broadcast(s[..., 0], o[..., 0]).shape + (LM_total_size(new_ell_min, new_ell_max),)
        new = np.zeros(new_shape, dtype=np.complex_)
        _multiplication_helper(s, self.ell_min, self.ell_max, self.s,
                               o, other.ell_min, other.ell_max, other.s,
                               new, new_ell_min, new_ell_max, new_s)
        metadata = copy.copy(self._metadata)
        metadata['spin_weight'] = new_s
        metadata['ell_min'] = new_ell_min
        metadata['ell_max'] = new_ell_max
        return type(self)(new, **metadata)
    else:
        if self._check_broadcasting(other):
            return self * other
        elif self._check_broadcasting(other, reverse=True):
            return other * self
        else:
            raise ValueError("Cannot broadcast input array to this Modes object.  Note that the input array\n           "
                             "must broadcast to all but last dimension of this object; it is not allowed to\n           "
                             "multiply each mode weight individually.  If you really want to hack this, view\n           "
                             "this Modes object as an ndarray, and don't complain when your results are wrong.")


def divide(self, other):
    if isinstance(other, type(self)):
        raise ValueError(f"Cannot divide one Modes object by another")
    else:
        return self / other



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

import copy
import math
import numpy as np


class Grid(np.ndarray):
    """Object to store SWHS values on a grid

    This class subclasses numpy's ndarray object, so that it should act like a numpy array in many
    respects, even with functions like np.zeros_like.

    NOTE: The functions `np.copy(grid)` and `np.array(grid, copy=True)` return `ndarray` objects;
    they lose information about the SWSH attributes, and have different ufuncs.  If you wish to keep
    this information, use `grid.copy()`.  Also note that pickling works as expected, as do copy.copy
    and copy.deepcopy.

    The number of dimensions is arbitrary as long as it is 2 or more, but the grid must be stored in
    the last 2 axes.  Specifically, the grid must have shape (n_theta, n_phi).  (See the function
    `spherical_functions.theta_phi` for an example of the actual grid locations expected.)  For
    example, a SWSH function of time may be stored as a 3-d array where the first axis represents
    different times, and the second and third axes represent the function values at each instant of
    time.

    This class also does two important things that are unlike numpy arrays:

    1) It tracks the spin weight of the function represented by this data.

    2) It overrides most of numpy's "universal functions" (ufuncs) to work appropriately for
       spin-weighted functions.  Specifically, these ufuncs are interpreted as acting on the
       spin-weighted function itself, rather than just the grid values.  The returned values are --
       where possible -- Grid objects.  Most importantly, we have these overriden methods:

       a) Multiplying two Grid objects will result in a new Grid object that represents the
          pointwise product of the functions themselves (and will correctly have spin weight given
          by the sum of the spin weights of the first two functions).  Division is only permitted
          when the divisor has spin weight zero.

       b) Addition (and subtraction) is permitted for functions of the same spin weight, but it does
          not make sense to add (or subtract) functions with different spin weights, so any attempt
          to do so raises a ValueError.

       Note that a constant is implicitly a function of spin weight zero, and is treated as such.

       Numerous other ufuncs -- such as log, exp, trigonometric ufuncs, bit-twiddling ufuncs, and so
       on -- are disabled because they don't result in objects of specific spin weights.

    It is possible to treat the underlying data of a Grid object `grid` as an ordinary numpy array
    by taking `grid.view(np.ndarray)`.  However, it is hoped that this class already performs all
    reasonable operations.  If you find a missing feature that requires you to resort to this,
    please feel free to open an issue in this project's github page to discuss it.

    Also, be aware that ndarrays also have various built-in methods that cannot be overridden very
    easily, such as max, copysign, etc.  If you try to use -- even indirectly -- those functions
    that don't have any clear interpretation for spin-weighted functions, things will likely break.


    Constructor parameters
    ======================
    input_array: array_like
        This may be a numpy ndarray, or any subclass.  It must be able to be viewed as complex.  If
        it has a `_metadata` attribute, that field will be copied to the new array; if the following
        parameters are not passed to the constructor, they will be searched for in the metadata.
    spin_weight: int [optional if present in `input_array._metadata`]
        The spin weight of the function that this Modes object describes.  This must be specified
        somehow, whether via a `_metadata` attribute of the input array, or as a keyword argument,
        or as the second positional argument (where the latter will override the former values).

    """

    # https://numpy.org/doc/1.18/user/basics.subclassing.html
    def __new__(cls, input_array, *args, **kwargs):
        if len(args) > 1:
            raise ValueError("Only one positional argument is recognized")
        elif len(args) == 1:
            kwargs['spin_weight'] = args[0]
        metadata = copy.copy(getattr(input_array, '_metadata', {}))
        metadata.update(**kwargs)
        input_array = np.asanyarray(input_array)
        if np.ndim(input_array) < 2:
            raise ValueError(f"Input array must have at least two dimensions; it has shape {input_array.shape}")
        n_theta, n_phi = input_array.shape[-2:]
        spin_weight = metadata.get('spin_weight', None)
        if spin_weight is None:
            raise ValueError("Spin weight must be specified")
        if n_theta < 2*abs(spin_weight)+1 or n_phi < 2*abs(spin_weight)+1:
            raise ValueError(f"Input array must have at least {2*abs(s)+1} points in each direction to have any "
                             f"nontrivial content for a field of spin weight {spin_weight}.")
        obj = input_array.view(cls)
        obj._metadata = metadata
        return obj

    # https://numpy.org/doc/1.18/user/basics.subclassing.html
    def __array_finalize__(self, obj):
        if obj is None: return
        self._metadata = copy.copy(getattr(obj, '_metadata', {}))
        if not 'spin_weight' in self._metadata:
            self._metadata['spin_weight'] = None

    # For pickling
    def __reduce__(self):
        state = super(Modes, self).__reduce__()
        new_attributes = state[2] + (self._metadata,)
        return (state[0], state[1], new_attributes)

    # For unpickling
    def __setstate__(self, state):
        self._metadata = copy.deepcopy(state[-1])
        super(Modes, self).__setstate__(state[:-1])

    @property
    def s(self):
        """Spin weight of this Modes object"""
        return self._metadata['spin_weight']

    spin_weight = s

    @property
    def n_theta(self):
        """Number of elements along the theta axis"""
        return self.shape[-2]

    @property
    def n_phi(self):
        """Number of elements along the phi axis"""
        return self.shape[-1]

    from .algebra import (
        conjugate, bar, real, imag, absolute,
        add, subtract, multiply, divide
    )

    conj = conjugate

    from .utilities import (
        modes, _check_broadcasting
    )

    from .ufuncs import __array_ufunc__



#!/usr/bin/env python

# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

from distutils.core import setup
from os import getenv

# Construct the version number from the date and time this python version was created.
from os import environ
from sys import platform
on_windows = ('win' in platform.lower() and not 'darwin' in platform.lower())
if "package_version" in environ:
    version = environ["package_version"]
    print("Setup.py using environment version='{0}'".format(version))
else:
    print("The variable 'package_version' was not present in the environment")
    try:
        # For cases where this is being installed from git.  This gives the true version number.
        from subprocess import check_output
        if on_windows:
            version = check_output("""git log -1 --format=%cd --date=format:'%Y.%m.%d.%H.%M.%S'""", shell=False)
            version = version.decode('ascii').strip().replace('.0', '.').replace("'", "")
        else:
            try:
                from subprocess import DEVNULL as devnull
                version = check_output("""git log -1 --format=%cd --date=format:'%Y.%-m.%-d.%-H.%-M.%-S'""", shell=True, stderr=devnull)
            except AttributeError:
                from os import devnull
                version = check_output("""git log -1 --format=%cd --date=format:'%Y.%-m.%-d.%-H.%-M.%-S'""", shell=True, stderr=devnull)
            version = version.decode('ascii').rstrip()
        print("Setup.py using git log version='{0}'".format(version))
    except Exception:
        # For cases where this isn't being installed from git.  This gives the wrong version number,
        # but at least it provides some information.
        #import traceback
        #print(traceback.format_exc())
        try:
            from time import strftime, gmtime
            try:
                version = strftime("%Y.%-m.%-d.%-H.%-M.%-S", gmtime())
            except ValueError:  # because Windows
                version = strftime("%Y.%m.%d.%H.%M.%S", gmtime()).replace('.0', '.')
            print("Setup.py using strftime version='{0}'".format(version))
        except:
            version = '0.0.0'
            print("Setup.py failed to determine the version; using '{0}'".format(version))
with open('spherical_functions/_version.py', 'w') as f:
    f.write('__version__ = "{0}"'.format(version))

validate = True
error_on_invalid = False
if getenv('CI') is not None:
    if getenv('CI').lower() == 'true':
        error_on_invalid = True

long_description = """\
Python/numba package for evaluating and transforming Wigner's 𝔇 matrices, Wigner's 3-j symbols, and
spin-weighted (and scalar) spherical harmonics. These functions are evaluated directly in terms of
quaternions, as well as in the more standard forms of spherical coordinates and Euler angles.
"""

if __name__ == "__main__":
    from setuptools import setup
    setup(name='spherical-functions',
          description='Python/numba implementation of Wigner D Matrices, spin-weighted spherical harmonics, and associated functions',
          long_description=long_description,
          url='https://github.com/moble/spherical_functions',
          author='Michael Boyle',
          author_email='mob22@cornell.edu',
          packages=[
              'spherical_functions',
              'spherical_functions.Wigner3j',
              'spherical_functions.WignerD',
              'spherical_functions.SWSH',
              'spherical_functions.SWSH_modes',
              'spherical_functions.SWSH_grids',
          ],
          package_data={'spherical_functions': ['*.npy']},
          version=version,
          install_requires=[
              'numpy>=1.13',
              'numpy-quaternion',
          ],
          python_requires='>=3.6',
    )



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

### NOTE: The functions in this file are intended purely for inclusion in the Grid class.  In
### particular, they assume that the first argument, `self` is an instance of Grid.  They should
### probably not be used outside of that class.


def modes(self, ell_max=None, **kwargs):
    """Return mode weights of function decomposed into SWSHs

    This method uses `spinsfast` to convert values on an equiangular grid to mode weights.

    The output array has one less dimension than this object; rather than the last two axes giving
    the values on the two-dimensional grid, the last axis gives the mode weights.

    Parameters
    ==========
    ell_max: None or int [defaults to None]
        Maximum ell value in the output.  If None, the result will have enough ell values to express
        the data on the grid without aliasing: (max(n_phi, n_theta) - 1) // 2.
    **kwargs: any types
        Additional keyword arguments are passed through to the Modes constructor on output

    """
    import copy
    import numpy as np
    import spinsfast
    from .. import Modes
    ell_max = ell_max or (max(n_phi, n_theta) - 1) // 2
    metadata = copy.copy
    return Modes(spinsfast.map2salm(self.view(np.ndarray), self.s, ell_max),
                 spin_weight=self.s, ell_min=0, ell_max=ell_max, **metadata)


def _check_broadcasting(self, array, reverse=False):
    """Test whether or not the given array can broadcast against this object"""
    import numpy as np

    if isinstance(array, type(self)):
        try:
            if reverse:
                np.broadcast(array, self)
            else:
                np.broadcast(self, array)
        except ValueError:
            return False
        else:
            return True
    else:
        if np.ndim(array) > np.ndim(self)-2:
            raise ValueError(f"Cannot broadcast array of {np.ndim(array)} dimensions against {type(self).__name__} "
                             f"object of fewer ({np.ndim(self)-2}) non-grid dimensions.\n"
                             "This is to ensure that scalars do not operate on individual "
                             "grid values; they must operate on all simultaneously.\n"
                             "If that is the case and you still want to broadcast, add more "
                             "dimensions before this object's first dimension.")
        try:
            if reverse:
                np.broadcast(array, self[..., 0, 0])
            else:
                np.broadcast(self[..., 0, 0], array)
        except ValueError:
            return False
        else:
            return True



# Copyright (c) 2019, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

"""Evaluating Wigner D matrices, spin-weighted spherical harmonics, and related.

This module contains code for evaluating the Wigner 3j symbols, the Wigner D
matrices, scalar spherical harmonics, and spin-weighted spherical harmonics.
The code is wrapped by numba where possible, allowing the results to be
delivered at speeds approaching or exceeding speeds attained by pure C code.

"""

from __future__ import print_function, division, absolute_import
from ._version import __version__

__all__ = ['Wigner3j', 'Wigner_D_element', 'Wigner_D_matrices', 'SWSH', 'SWSH_grid',
           'factorial', 'binomial_coefficient', 'ladder_operator_coefficient']

import numpy as np
from math import factorial
import os.path

from quaternion.numba_wrapper import njit, xrange

# Module constants
ell_max = 32  # More than 29, and you get roundoff building quickly
epsilon = 1.e-15
error_on_bad_indices = True

# The coefficient files
#   binomial_coefficients.npy
#   ladder_operator_coefficients.npy
#   Wigner_coefficients.npy
# were originally produced with the code in `_generate_coefficients.py`.


# Factorial
factorials = np.array([float(factorial(i)) for i in range(171)])


@njit('f8(i8)')
def factorial(i):
    return factorials[i]


# Binomial coefficients
_binomial_coefficients = np.load(os.path.join(os.path.dirname(__file__), 'binomial_coefficients.npy'))

@njit('f8(i8,i8)')
def binomial_coefficient(n, k):
    return _binomial_coefficients[(n * (n + 1)) // 2 + k]


# Ladder-operator coefficients
_ladder_operator_coefficients = np.load(os.path.join(os.path.dirname(__file__), 'ladder_operator_coefficients.npy'))

@njit('f8(i8,i8)')
def _ladder_operator_coefficient(twoell, twom):
    return _ladder_operator_coefficients[((twoell + 2) * twoell + twom) // 2]

@njit('f8(f8,f8)')
def ladder_operator_coefficient(ell, m):
    return _ladder_operator_coefficient(round(2*ell), round(2*m))


# Coefficients used in constructing the Wigner D matrices
_Wigner_coefficients = np.load(os.path.join(os.path.dirname(__file__), 'Wigner_coefficients.npy'))

@njit('i8(i8,i8,i8)')
def _Wigner_index(twoell, twomp, twom):
    return twoell*((2*twoell + 3)*twoell + 1) // 6 + (twoell + twomp)//2 * (twoell + 1) + (twoell + twom)//2

@njit('f8(i8,i8,i8)')
def _Wigner_coefficient(twoell, twomp, twom):
    return _Wigner_coefficients[_Wigner_index(twoell, twomp, twom)]

@njit('f8(f8,f8,f8)')
def Wigner_coefficient(ell, mp, m):
    return _Wigner_coefficient(round(2*ell), round(2*mp), round(2*m))


def LM_range(ell_min, ell_max):
    """Array of (ell,m) indices in standard order

    This function returns an array of essentially

    [[ell,m] for ell in range(ell_min, ell_max+1)
             for m in range(-ell,ell+1)]

    This is, for example, the order assumed for mode data in the `waveforms`
    module.

    """
    # # Sympy commands to calculate the total size:
    # from sympy import symbols, summation
    # ell_min,ell,ell_max = symbols('ell_min,ell,ell_max', integer=True)
    # summation((2*ell + 1), (ell, ell_min, ell_max))
    LM = np.empty((ell_max * (ell_max + 2) - ell_min ** 2 + 1, 2), dtype=int)
    _LM_range(ell_min, ell_max, LM)
    return LM


@njit('void(i8,i8,i8[:,:])')
def _LM_range(ell_min, ell_max, LM):
    i = 0
    for ell in xrange(ell_min, ell_max + 1):
        for m in xrange(-ell, ell + 1):
            LM[i, 0] = ell
            LM[i, 1] = m
            i += 1


@njit('i8(i8,i8,i8)')
def LM_index(ell, m, ell_min):
    """Array index for given (ell,m) mode

    Assuming an array of

    [[ell,m] for ell in range(ell_min, ell_max+1)
             for m in range(-ell,ell+1)]

    this function returns the index of the (ell,m) element.  (Note that
    ell_max doesn't actually come into this calculation, so it is not taken
    as an argument to the function.)

    This can be calculated in sympy as

      from sympy import symbols, summation
      ell,m,ell_min, = symbols('ell,m,ell_min,', integer=True)
      summation(2*ell + 1, (ell, ell_min, ell-1)) + (ell+m)

    """
    return ell * (ell + 1) - ell_min ** 2 + m


@njit('i8(i8,i8)')
def LM_total_size(ell_min, ell_max):
    """Total array size of (ell,m) components

    Assuming an array of

    [[ell,m] for ell in range(ell_min, ell_max+1)
             for m in range(-ell,ell+1)]

    this function returns the total size of that array.

    This can be calculated in sympy as

      from sympy import symbols, summation
      ell,ell_min,ell_max = symbols('ell,ell_min,ell_max', integer=True)
      summation(2*ell + 1, (ell, ell_min, ell_max))

    """
    return ell_max * (ell_max + 2) - ell_min ** 2 + 1


def LM_deduce_ell_max(size, ell_min=0):
    ell_max = int(np.sqrt(size + ell_min**2) - 1)
    if ell_max * (ell_max + 2) - ell_min ** 2 + 1 != size:
        raise ValueError(f"The input size {size} does not correspond to a possible array of modes with ell_min {ell_min}.")
    return ell_max


def LMpM_range(ell_min, ell_max):
    """Array of (ell,mp,m) indices in standard order

    This function returns an array of essentially

    [[ell,mp,m] for ell in range(ell_min, ell_max+1)
                for mp in range(-ell,ell+1)
                for m in range(-ell,ell+1)]

    This is, for instance, the array of indices of the Wigner D matrices
    constructed by this module.

    """
    # # Sympy commands to calculate the total size:
    # from sympy import symbols, summation
    # ell_min,ell,ell_max = symbols('ell_min,ell,ell_max', integer=True)
    # summation((2*ell + 1)**2, (ell, ell_min, ell_max))
    if abs(round(ell_max)-ell_max) > 1e-10 or abs(round(ell_min)-ell_min) > 1e-10:
        error = ("LMpM_range is only intended for integer values of ell.\n"
                 + "Input values ell_min={0} and ell_max={1} are not valid.\n".format(ell_min, ell_max)
                 + "Try `LMpM_range_half_integer` instead.")
        raise ValueError(error)
    LMpM = np.empty(((ell_max * (11 + ell_max * (12 + 4 * ell_max)) + ell_min * (1 - 4 * ell_min ** 2) + 3) // 3, 3),
                    dtype=int)
    _LMpM_range(ell_min, ell_max, LMpM)
    return LMpM

@njit('void(i8,i8,i8[:,:])')
def _LMpM_range(ell_min, ell_max, LMpM):
    i = 0
    for ell in xrange(ell_min, ell_max + 1):
        for mp in xrange(-ell, ell + 1):
            for m in xrange(-ell, ell + 1):
                LMpM[i, 0] = ell
                LMpM[i, 1] = mp
                LMpM[i, 2] = m
                i += 1


def LMpM_range_half_integer(ell_min, ell_max):
    """Array of (ell,mp,m) indices in standard order, including half-integer values

    This function returns an array of essentially

    [[twoell/2, twomp/2, twom/2]
     for twoell in range(2*ell_min, 2(ell_max+1)
     for twomp in range(-twoell, twoell+1, 2)
     for twom in range(-twoell, twoell+1, 2)]

    See also: LMpM_range

    """
    # # Sympy commands to calculate the total size:
    # from sympy import symbols, summation
    # twoell_min,twoell,twoell_max = symbols('twoell_min,twoell,twoell_max', integer=True)
    # summation((twoell + 1)**2, (twoell, twoell_min, twoell_max))
    LMpM = np.empty((int(((8*ell_max + 18)*ell_max + 13)*ell_max + 3 - ((8 * ell_min + 6) * ell_min + 1)*ell_min) // 3,
                     3), dtype=float)
    _LMpM_range_half_integer(round(2*ell_min), round(2*ell_max), LMpM)
    return LMpM

@njit('void(i8,i8,f8[:,:])')
def _LMpM_range_half_integer(twoell_min, twoell_max, LMpM):
    i = 0
    for twoell in xrange(twoell_min, twoell_max + 1):
        for twomp in xrange(-twoell, twoell + 1, 2):
            for twom in xrange(-twoell, twoell + 1, 2):
                LMpM[i, 0] = twoell / 2
                LMpM[i, 1] = twomp / 2
                LMpM[i, 2] = twom / 2
                i += 1


@njit('i8(i8,i8,i8,i8)')
def LMpM_index(ell, mp, m, ell_min):
    """Array index for given (ell,mp,m) mode

    Assuming an array (e.g., Wigner D matrices) in the order

    [[ell,mp,m] for ell in range(ell_min, ell_max+1)
                for mp in range(-ell,ell+1)
                for m in range(-ell,ell+1)]

    this function returns the index of the (ell,mp,m) element.  (Note that
    ell_max doesn't actually come into this calculation, so it is not taken
    as an argument to the function.)

    This can be calculated in sympy as

      from sympy import symbols, summation
      ell,mp,m,ell_min, = symbols('ell,mp,m,ell_min,', integer=True)
      summation((2*ell + 1)**2, (ell, ell_min, ell-1)) + (2*ell+1)*(ell+mp) + (ell+m)

    """
    # raw output is: 4*ell**3/3 + 2*ell**2 + 2*ell*mp + 5*ell/3 - 4*ell_min**3/3 + ell_min/3 + m + mp
    # We rearrange that to act more nicely
    return (((4 * ell + 6) * ell + 6 * mp + 5) * ell + ell_min * (1 - 4 * ell_min ** 2) + 3 * (m + mp)) // 3


@njit('i8(i8, i8)')
def LMpM_total_size(ell_min, ell_max):
    """Total array size of Wigner D matrix

    Assuming an array (e.g., Wigner D matrices) in the order

    [[ell,mp,m] for ell in range(ell_min, ell_max+1)
                for mp in range(-ell,ell+1)
                for m in range(-ell,ell+1)]

    this function returns the total size of that array.

    This can be calculated in sympy as

      from sympy import symbols, summation
      ell,ell_min,ell_max = symbols('ell,ell_min,ell_max', integer=True)
      summation((2*ell + 1)**2, (ell, ell_min, ell_max))

    """
    # raw output is: 4*ell_max**3/3 + 4*ell_max**2 + 11*ell_max/3 - 4*ell_min**3/3 + ell_min/3 + 1
    # We rearrange that to act more nicely
    return (((4 * ell_max + 12) * ell_max + 11) * ell_max + (-4 * ell_min ** 2 + 1) * ell_min + 3) // 3


def theta_phi(n_theta, n_phi):
    """Construct (theta, phi) grid

    This grid is in the order expected by spinsfast

    Parameters
    ==========
    n_theta: int
        Number of points in the theta direction
    n_phi: int
        Number of points in the phi direction

    Returns
    =======
    theta_phi_grid: ndarray
        Array of pairs of floats giving the respective [theta, phi] pairs.  The shape of this array
        is (n_theta, n_phi, 2).

    """
    return np.array([[[theta, phi]
                      for phi in np.linspace(0.0, 2*np.pi, num=n_phi, endpoint=False)]
                     for theta in np.linspace(0.0, np.pi, num=n_theta, endpoint=True)])


from .Wigner3j import Wigner3j, clebsch_gordan
from .WignerD import (Wigner_D_element, _Wigner_D_element,
                      Wigner_D_matrices, _Wigner_D_matrices,
                      _linear_matrix_index, _linear_matrix_diagonal_index,
                      _linear_matrix_offset, _total_size_D_matrices)
from .SWSH import SWSH, SWSH_grid, _SWSH  # sYlm, Ylm
from .SWSH_modes import Modes
from .SWSH_grids import Grid
from .mode_conversions import (constant_as_ell_0_mode, constant_from_ell_0_mode,
                               vector_as_ell_1_modes, vector_from_ell_1_modes,
                               eth_GHP, ethbar_GHP, eth_NP, ethbar_NP,
                               ethbar_inverse_NP)
from .multiplication import multiply



import math
import numpy as np
from . import LM_total_size, Wigner3j, LM_index
from quaternion.numba_wrapper import jit, njit, xrange


@njit()
def _multiplication_helper(f, ellmin_f, ellmax_f, s_f,
                           g, ellmin_g, ellmax_g, s_g,
                           fg, ellmin_fg, ellmax_fg, s_fg):
    # NOTE: This works in-place on fg, and returns fg for good measure
    for ell1 in range(ellmin_f, ellmax_f+1):
        for m1 in range(-ell1, ell1+1):
            sum1 = math.sqrt((2*ell1+1)/(4*math.pi))*f[..., LM_index(ell1, m1, ellmin_f)]  # Calculate f contribution
            for ell2 in range(ellmin_g, ellmax_g+1):
                for m2 in range(-ell2, ell2+1):
                    sum2 = math.sqrt(2*ell2+1)*g[..., LM_index(ell2, m2, ellmin_g)]   # Calculate g contribution
                    m3 = m1+m2
                    for ell3 in range(max(abs(m3), abs(ell1-ell2)), min(ell1+ell2, ellmax_fg)+1):
                        # Could loop over same (ell3, m3) more than once, so add all contributions together
                        fg[..., LM_index(ell3, m3, ellmin_fg)] += (
                            math.pow(-1, ell1 + ell2 + ell3 + s_fg + m3)
                            * math.sqrt(2*ell3+1)
                            * Wigner3j(ell1, ell2, ell3, s_f, s_g, -s_fg)
                            * Wigner3j(ell1, ell2, ell3, m1, m2, -m3)
                        ) * sum1 * sum2
    return fg


def multiply(f, ellmin_f, ellmax_f, s_f, g, ellmin_g, ellmax_g, s_g):
    """Return modes of the decomposition of f*g

    s1Yl1m1 * s2Yl2m2 = sum([
        s3Yl3m3.conjugate() * (-1)**(l1+l2+l3) * sqrt((2*l1+1)*(2*l2+1)*(2*l3+1)/(4*pi))
            * Wigner3j(l1, l2, l3, s1, s2, s3) * Wigner3j(l1, l2, l3, m1, m2, m3)
        for l3 in range(abs(l1-l2), l1+l2+1)
    ])

    Here, s3 = -s1 - s2 and m3 = -m1 - m2.

    For general f and g with random ellmin/max and m:
    f*g = sum([ sqrt(3*(2*l1+1)/4*pi)*f(l1,m1)
               sum([ sqrt(2*l2+1)*g(l2,m2)
                    sum([ s3Yl3m3.conjugate()*sqrt(2*l3+1)*(-1)**(l1+l2+l3)
                          *Wigner3j(l1, l2, l3, s1, s2, s3) * Wigner3j(l1, l2, l3, m1, m2, m3)
                    for l3 in range(abs(l1-l2), l1+l2+1)
                    ])
               for l2,m2 in range(ellmin_g, ellmax_g)
               ])
          for l1,m1 in range(ellmin_f, ellmax_f)
          ])
    for same s3,m3 as above


    Parameters
    ----------
    f: complex array
        This gives the mode weights of the function `f` expanded in spin-weighted spherical
        harmonics.  They must be stored in the standard order:
            [f(ell, m) for ell in range(ellmin_f, ellmax_f+1) for m in range(-ell, ell+1)]
        In particular, it is permissible to have ellmin < |s|, even though any coefficients for such
        values should be zero; they will just be ignored.
    ellmin_f: int
        See `f`
    ellmax_f: int
        See `f`
    s_f: int
        Spin weight of `f` function
    g: complex array
        As above, but for the second function
    ellmin_g: int
        As above, but for the second function
    ellmax_f: int
        As above, but for the second function
    s_f: int
        As above, but for the second function

    Returns
    -------
    fg: complex array
    ellmin_fg: int
    ellmax_fg: int
    s_fg: int
        These outputs are analogous to the inputs, except for f*g.  The spin weight is necessarily
            s_fg = s_f + s_g
        and the max ell value is necessarily
            ellmax_fg = ellmax_f + ellmax_g
        For simplicity, we always return ellmin_fg=0, even though this may not be necessary.

    """
    s_fg = s_f + s_g
    ellmax_fg = ellmax_f + ellmax_g
    ellmin_fg = 0
    shape_fg = np.broadcast(f[..., 0], g[..., 0]).shape + (LM_total_size(0, ellmax_fg),)
    fg = np.zeros(shape_fg, dtype=np.complex_)

    _multiplication_helper(f, ellmin_f, ellmax_f, s_f,
                           g, ellmin_g, ellmax_g, s_g,
                           fg, ellmin_fg, ellmax_fg, s_fg)

    return fg, ellmin_fg, ellmax_fg, s_fg
